/*------------------------------------------------------------------------
--                                                                      --
--          H E L I O S   I N P U T / O U T P U T   S E R V E R         --
--          ---------------------------------------------------         --
--                                                                      --
--               Copyright (C) 1987, Perihelion Software Ltd.           --
--                             All Rights Reserved.                     --
--                                                                      --
--  Devices.c                                                           --
--                                                                      --
--           This module contains code to interact with the various     --
--                                                                      --
--           devices: clock, printer, rs232, centronics,                --
--                                                                      --
--           mouse, and raw keyboard. It also contains some polling     --
--                                                                      --
--           code                                                       --
--                                                                      --
--  Author:  BLV 30/11/87                                               --
--                                                                      --
------------------------------------------------------------------------*/
/* SccsId: 3.7 14/4/89\ Copyright (C) 1987, Perihelion Software Ltd.        */

/**
*** The usual header file and some function declarations.
**/

#define Devices_Module
#include "helios.h"

FORWARD void fn( add_key,          (int, Window *));
PRIVATE void fn( send_break_event, (Port));
PRIVATE void fn( copy_event,       (IOEvent *, IOEvent *));


/*------------------------------------------------------------------------
--------------------------------------------------------------------------
--  I start off with the various polling routines.                      --
--  My usual approach is to put data into circular buffers when         --
--  handling interrupts, and extract them here. Poll_the_devices is     --
--  called from within the main loop of the server.                     --
--------------------------------------------------------------------------
------------------------------------------------------------------------*/

/**
*** Routine poll_the_devices() is called from inside the Server's and
*** from the debugger's main loop. The purpose of the routine is to examine the
*** status of all the devices that need checking and take whatever action
*** is appropriate. This may involve sending mouse and keyboard events to
*** the transputer, reading the keyboard for each window, polling Gem,
*** and checking some internal flags.
**/


void poll_the_devices()
{ PRIVATE void fn( poll_the_windows,  (void));

#if keyboard_supported
  PRIVATE int  fn( poll_the_keyboard, (void));

  if (poll_the_keyboard())
#endif
     poll_the_windows();

#if mouse_supported
  { PRIVATE void fn( poll_the_mouse, (void));
    poll_the_mouse();
  }
#endif

#if gem_supported
    poll_gem();
#endif

#if RS232_supported
  { if (RS232_errno)
     RS232_check_events();
    RS232_errno = 0;
  }
#endif

}

PRIVATE void poll_the_windows()
{ int temp;

  while ((temp = read_char_from_keyboard(Server_window.handle) ) ne -1)
    add_key(temp, &Server_window);

#if multiple_windows
   { int temp;
     Window *window;
     for (window = (Window *) Window_List.list.head;
          window->node.node.next ne (Node *) NULL;
          window = (Window *) window->node.node.next )
       while ((temp = read_char_from_keyboard(window->handle)) ne -1)
         add_key(temp, window);
   }
#endif /* multiple_windows */
}

void RS232_send_event(port, event_type)
WORD port;
WORD event_type;
{ 
  IOEvent *event = (IOEvent *) mcb->Data;

  event->Type    = event_type;
  event->Counter = 0;
  event->Stamp   = Now;

  copy_event(event, event);   /* This does all the swapping and conversions */

  mcb->MsgHdr.Reply = port;
  mcb->MsgHdr.Dest  = (word) NULL;
  mcb->MsgHdr.Flags = MsgHdr_Flags_preserve;

  Request_Return(EventRc_IgnoreLost, 0L, (word) sizeof(IOEvent));
}


/**
*** This routine is used to put characters typed at the keyboard into the
*** console input buffer. It takes care of certain special keys, e.g.
*** ctrl-C, ctrl-S and ctrl-Q.
***
*** Adding something to a circular buffer - the something is added to the head
*** of the table, the head is increased modulo the limit, and then there is
*** a check for overflow, i.e. has the entire table been filled. There are
*** two possible approaches here, either forget about the bit of data just
*** added or forget about the oldest bit of data, and different approaches
*** are suitable for different devices. For the console, if somebody has
*** typed 256 characters ahead I prefer to lose the newest bit of data.
**/

void add_key(key, window)
int key;
Window *window;
{ switch(key)
   {
     case 0x03 :        /* ctrl-C */
                        /* if breaks are ignored, ignore ctrl-C */
                if (IsAnAttribute(&(window->attr), ConsoleIgnoreBreak))
                  return;
                else if /* is the event handler active ? */
                 (IsAnAttribute(&(window->attr), ConsoleBreakInterrupt))
                 { if (window->break_handler.port ne (WORD) NULL)
                     send_break_event(window->break_handler.port);
                   if (window->cooker.readerco ne (Conode *) NULL)
                    { (window->cooker.readerco)->flags |= CoFlag_CtrlCed;
                      (window->cooker.readerco)->type = CoReady;
                    }
                   return;
                 }

                break;  /* otherwise put it in the buffer */

      case 0x11 :       /* ctrl-Q */
                 if (window->XOFF)
                  { window->XOFF--; return; }
                 else
                  break;

      case 0x13 :      /* ctrl-S */
                 if (IsAnAttribute(&(window->attr), ConsolePause))
                  { window->XOFF++; return; }
                 else
                  break;
        }


  window->Table[window->head] = (UBYTE) key;
  window->head = (window->head + 1) & (Console_limit - 1);
  if (window->head eq window->tail)   /* buffer overflow */
    window->head = (window->head - 1) & (Console_limit - 1); 
}


/**
*** When a break event has been detected, the console attributes are set such
*** that they are not ignored and are not turned into ordinary keys, and a
*** break_handler has been installed, the routine below is called to send off
*** the event. The event is set up, copied onto itself to take care of
*** swapping and time stamp conversion, and sent off. The FnRc is set to
*** IgnoreLost because I do not care about losing break events - if the user
*** detects that the break event has not worked he will just press the key
*** again.
**/

PRIVATE void send_break_event(port)
Port port;
{ 
  IOEvent *event = (IOEvent *) mcb->Data;

  event->Type    = Event_Break;
  event->Counter = 0;
  event->Stamp   = Now;

  copy_event(event, event);   /* This does all the swapping and conversions */

  mcb->MsgHdr.Reply = port;
  mcb->MsgHdr.Dest  = (word) NULL;
  mcb->MsgHdr.Flags = MsgHdr_Flags_preserve;

  Request_Return(EventRc_IgnoreLost, 0L, (word) sizeof(IOEvent));
}


/**
*** The following code is very useful for taking an event such as a mouse
*** or keyboard event and putting it into a message data area, because it
*** does all the swapping required and does the time stamp conversion.
*** NB it is possible for the copy/conversion to happen in place.
*** The time stamp is supposed to be in milliseconds since start-up, but
*** the stamp I store at the moment comes from Now.
**/

PRIVATE void copy_event(to, from)
IOEvent *to, *from;
{ to->Type     = swap(from->Type);
  to->Counter  = swap(from->Counter);
  to->Stamp    =
   swap(divlong( ((from->Stamp - initial_stamp) * 1000L), (WORD) CLK_TCK) );
  to->Device.Break.junk1 = swap(from->Device.Break.junk1);
  to->Device.Break.junk2 = swap(from->Device.Break.junk2);
}


/*-----------------------------------------------------------------------------
-------------------------------------------------------------------------------
--
--    The following routines deal with setting up the devices. Initialise is --
-- called during startup only, and restore devices is called when the server --
-- exits. Restart is called during a reboot and during start-up              --
--
------------------------------------------------------------------------------- 
-----------------------------------------------------------------------------*/

/**
*** The code for this tends to be rather specialised so I prefer to put it
*** into the hardware-specific files, ST/STLOCAL.C etc.
*** It is important that all the event handlers are cleared each time the
*** Server exits or reboots.
***
*** I am not certain what to do about data still in the circular buffers
*** when the user wishes to reboot, but scrapping it is probably sensible.
**/
PRIVATE int devices_initialised = 0;

void initialise_devices()
{ devices_initialised = 1;

#if multiple_windows
/**
*** The window_list is initialised here as well as in Window_InitServer(),
*** just in case...
**/
  InitList(&Window_List.list);
  Window_List.entries = 0;

/**
*** The Server_window is used mainly to display the Server's debugging
*** info, and as the window for debugging mode. Hence I set the
*** Server_window to a raw-ish mode. If multiple windows are not supported
*** then Console_InitServer is responsible for initialising all of the
*** Server_window structure.
**/
  Server_window.handle = create_a_window("Helios Server");
  if (Server_window.handle eq 0L)
   { printf("Unable to create a window for the Server.\n");
     longjmp(exit_jmpbuf, 1);
   }

  strcpy(&(Server_window.node.direntry.Name[0]), "Helios Server");
  Server_window.XOFF   = 0;
  Server_window.head   = 0;
  Server_window.tail   = 0;
  InitAttributes(&(Server_window.attr));
  RemoveAttribute   (&(Server_window.attr), ConsoleEcho);
  RemoveAttribute(&(Server_window.attr), ConsoleIgnoreBreak);
  RemoveAttribute   (&(Server_window.attr), ConsoleBreakInterrupt);
  RemoveAttribute   (&(Server_window.attr), ConsolePause);

#if use_ANSI_emulator
  { word x, y;
    window_size(Server_window.handle, &x, &y);
    
    if (!Init_Ansi(&Server_window, (int) y, (int) x))
     { printf("Unable to initialise ANSI emulator library for server window.\n");
       longjmp(exit_jmpbuf, 1);
     }
  }
#endif /* use_ANSI_emulator */

#else  /* not multiple windows */
#if use_ANSI_emulator
    if (!Init_Ansi(&Server_window,
#if (PC || ST)
        25, 80))
#endif
     { printf("Unable to  initialise ANSI emulator library for server window.\n");
       longjmp(exit_jmpbuf, 1);
     }
#endif /* use_ANSI_emulator */

#endif /* multiple_windows */

#if ST
  ST_initialise_devices();
#endif
#if PC
  { PUBLIC void fn( PC_initialise_devices, (void));
    PC_initialise_devices();
  }
#endif
#if AMIGA
  Amiga_initialise_devices();
#endif
#if SUN
  Sun_initialise_devices();
#endif
}

void restore_devices()
{ if (!devices_initialised) return;
#if ST
  ST_restore_devices();
#endif
#if PC
  { PUBLIC void fn( PC_restore_devices, (void));
    PC_restore_devices();
  }
#endif
#if AMIGA
  Amiga_restore_devices();
#endif
#if SUN
  Sun_restore_devices();
#endif

#if multiple_windows
  close_window(Server_window.handle);
#endif
}

void restart_devices()
{
#if RS232_supported
  RS232_errno = 0;
#endif

#if Centronics_supported
  Centronics_errno = 0;
#endif

#if Printer_supported
  Printer_errno = 0;
#endif

#if floppies_available
  floppy_errno = 0;
#endif

#if Midi_supported
  Midi_errno = 0;
#endif

#if Gem_supported
  restart_gem();
#endif
}

/*******************************************************************************
********************************************************************************
********************************************************************************
********************************************************************************
********************************************************************************
********* That is all the general purpose stuff. The code below deals with *****
********* the devices actually supported.                                  *****
********************************************************************************
********************************************************************************
********************************************************************************
********************************************************************************
*******************************************************************************/

/*------------------------------------------------------------------------
--------------------------------------------------------------------------
--
-- The Clock device
--
--------------------------------------------------------------------------
------------------------------------------------------------------------*/

/**
*** The clock is a very simple device to implement because the only valid
*** requests are Locate and Create, which can be done by default handlers,
*** ObjectInfo, which is used by the transputer to discover the current
*** time, and SetDate to allow the transputer to change the time.
***
*** ObjectInfo is straightforward because most of the fields in the ObjInfo
*** structure are useless, and I only need to worry about the date stamps
*** which are the system startup time for creation and the current time for
*** last altered and accessed. It is useful to have some way of finding out
*** when the system started up, and clock creation seems the best way of doing
*** it. To get hold of the current time I use the rawest system clock available
*** thus bypassing any conversions, e.g. timezone ones, done by the C library.
**/

void Clock_ObjectInfo(myco)
Conode *myco;
{                      /* put info in data vector straightaway, nothing in */
                       /* there that we need any more */
  ObjInfo *Heliosinfo = (ObjInfo *) mcb->Data;
  WORD current_time;
  if (strcmp(IOname, "clock") )
    { Request_Return(EC_Error + SS_IOProc + EG_Name + EO_Message, 0L, 0L);
      return;
    }   

  current_time = get_unix_time();

  Heliosinfo->DirEntry.Type   = swap(Type_Private);
  Heliosinfo->DirEntry.Flags  = swap(0L);
  Heliosinfo->DirEntry.Matrix = swap(DefDirMatrix);
  strcpy(Heliosinfo->DirEntry.Name, "clock");
  Heliosinfo->Account   = swap(0L);
  Heliosinfo->Size      = swap(0L);
  Heliosinfo->Creation  = swap(Startup_Time);
  Heliosinfo->Access    = swap(current_time);
  Heliosinfo->Modified  = Heliosinfo->Access;

  Request_Return(ReplyOK, 0L, (word) sizeof(ObjInfo));
  use(myco)
}

/**
*** Setting the system clock is not possible on all machines, e.g. under Unix
*** only the super user is allowed to change the clock.
**/

#if !(SUN)

void Clock_SetDate(myco)
Conode *myco;
{ WORD unixstamp = (mcb->Control)[SetDateDate_off];
  extern void fn( set_current_time, (WORD));

  set_current_time(unixstamp);

  Request_Return(ReplyOK, 0L, 0L);
  use(myco)
}

#endif   /* Clock settable */


/*------------------------------------------------------------------------
--------------------------------------------------------------------------
--
--     The Host device
--
--------------------------------------------------------------------------
------------------------------------------------------------------------*/
/**
*** The Host device is one I do not like very much, but all our customers
*** seem to want it. Its purpose is to provide limited (or maybe not so
*** limited) interaction between Helios and the machine running the host.
*** Each host has a device /pc or /st or whatever that can be opened, and
*** once it is open it can be sent private protocol messages. These result
*** in some machine-dependant operation, followed by a return message.
*** On the PC and the ST the message results in a trap being invoked. On
*** other systems, particularly multi-tasking systems, it might be possible
*** to provide full bi-directional communication between Helios and host
*** programs, though I have no intention of writing the relevant code myself.
**/

#if interaction_supported

void Host_Open(myco)
Conode *myco;
{          	                             /* check the device we are opening */
  if (strcmp(IOname, machine_name) )
    { Request_Return(EC_Error + SS_IOProc + EG_Name + EO_Message, 0L, 0L);
      return;
    }

  NewStream(Type_File, Flags_Closeable + Flags_Interactive, NULL, 
            Host_Handlers);
  use(myco)
}

#if (ST || PC)              /* The behaviour varies from machine to machine */
void Host_PrivateStream(myco)
Conode *myco;
{ WORD   reply_port = mcb->MsgHdr.Reply;    /* Save the reply port */
  WORD   reply;
  extern WORD fn( call_a_trap, (MCB *));   /* The machine dependant routine */
                                           /* Probably written in assembler */
  reply = call_a_trap(mcb);

  mcb->MsgHdr.Reply = reply_port;
  Request_Return(reply, (WORD) mcb->MsgHdr.ContSize,
                 (WORD) mcb->MsgHdr.DataSize);
  use(myco)
}
#endif  /* ST || PC */

void Host_Close(myco)
Conode *myco;
{ if (mcb->MsgHdr.Reply ne NullPort)
   Request_Return(ReplyOK, 0L, 0L);
  Seppuku();
  use(myco)
}

#endif  /* interaction_supported */

/*------------------------------------------------------------------------
--
-- Console device - this has been moved to terminal.c . It had got out of
-- hand when I added support for multiple windows.
--
------------------------------------------------------------------------*/
/*------------------------------------------------------------------------
--
-- Port IO
--
------------------------------------------------------------------------*/


/**
*** All the various communication ports, RS232, Centronics, Midi etc., tend
*** to work in much the same way, so I use the same code for most of them.
*** In addition, I treat Printer devices, which may or may not use similar
*** communication stuff, in the same way.
***
*** To implement a port for your hardware is relatively easy. First, you must
*** change the header file defines.h for your machine. Next you need to write
*** a small number of routines : x_initlist(), x_send(), x_pollwrite(),
*** x_pollread(), x_receive(), x_pollread(), x_abortread(), x_configure(), and
*** x_done(). These are all described in detail below. You also need to use the
*** error mechanism involving x_errno. 
***
*** The main problem with these devices is that there may be large numbers of
*** them. For example, a humble PC can have upto 7 serial and 3 parallel ports,
*** and there is no limit on the number of printers that may be hooked up to
*** a Unix or Vax installation. Hence all the devices are treated as directories
*** and each entry in the directory corresponds to a physical port or device.
*** Thus to produce a printout you might write to /Printers/Laser, or to
*** /Centronics/Port0. I have no way of knowing what entries you want in these
*** directories, so your initlist() routine must put all the entries in a linked
*** list for me. Each entry consists of a Port_node structure as declared in
*** structs.h, and you are responsible for filling in the direntry.Name and the
*** id. To make this easier I provide an add_port_node() routine below.
***
*** It is extremely desirable to have some common name on all the bits of
*** hardware. Hence I put a Default entry in each directory, with an id of 0.
*** This allows programs to output to e.g. /Printers/Default if no other
*** printer is specified. Your code has to allow for this id 0 as a special
*** case.
***
*** Given the list of Port_node's, I can deal with all the nasty bits like
*** listing the device directory. However, I have no way of knowing how
*** to do physical IO with the devices, so you must provide the necessary
*** routines. These are a start routine to set up the IO, a poll routine to
*** see how the IO is progressing, and an abort routine to terminate the IO
*** if it times out. It is up to you how you implement these. You may start up
*** some interrupt-driven code in the start routine, and just check a flag in
*** the poll routine to see whether or not the IO has finished. Alternatively
*** you may have the poll routine doing all the hard work, i.e. check the
*** device status and do IO if the device is ready.
***
*** There are two other routines : a done function and a configure function.
*** The done function is called when the stream is closed, either explicitly
*** or due to a stream timeout. If you access say a printer by writing to a
*** pseudo-file and closing it, with the system doing ALL the buffering and
*** never suspending you because it runs out of buffers, then you might
*** implement a device by having the start function do the write, and having
*** the poll function succeed straight away - in the done function you close
*** the file to start the printer. The configure function is used to change the
*** attributes of a device, in response to a SetAttributes() call and during
*** the start-up of some devices.
***
*** The code below. add_port_node() is called by the local.c modules to add
*** entries to the device directories, and it is also called by the InitServer()
*** routines. TidyServer() is responsible for freeing the space taken up by the
*** list of devices. Note that the list header is a global, so that it should
*** not be freed. Locate, ObjInfo and Open are fairly straightforward : the only
*** problem lies in detecting whether the request refers to the directory or to
*** a device inside the directory, and if the latter to which device. I use a
*** find_node routine below to find the particular device. The code for
*** accessing the device directory is shared with that for file directories
*** and for the IOprocessor device, via fundefs.h, so no extras are needed for
*** that. This leaves the code for dealing with device streams. These are
*** implemented in terms of the device-specific functions, i.e. x_pollwrite()
*** etc. and they are fairly straightforward. Both Read's and Write's go via the
*** same General_PortIO routine.
***
*** Quick comment on the extra field in the coroutine nodes. The extra field in
*** the server node points to a directory header, which is a list of all the
*** Port_node's plus a count of the number of nodes. The extra field for the
*** directory stream is exactly the same, since it is in the required format for
*** the directory routines. The extra field for a device stream points to the
*** Port_node for that device, which includes amongst other things the device
*** functions and its id.
***
*** This comment is getting too big, so I shall leave the device routines to
*** the RS232 section below.
**/

#if Ports_used

#define Port_read   1L
#define Port_write  2L
PRIVATE void fn( General_PortIO, (WORD, WORD, Conode *, WordFnPtr, WordFnPtr, WordFnPtr, VoidFnPtr));
PRIVATE Port_node *fn( find_node, (Conode *));

void add_port_node(list, name, id, attr)
List *list;
char *name;
WORD id;
Attributes *attr;
{ Port_node *node = malloc(sizeof(Port_node));
  if (node eq (Port_node *) NULL)
    { ServerDebug("Insufficient memory to initialise devices.");
      longjmp(exit_jmpbuf, 1);
    }

  strcpy(node->direntry.Name, name);
  node->direntry.Type = Type_File;
  node->id = id;
  node->attr = attr;
  AddTail(&(node->node), list);
}

void Port_TidyServer(myco)
Conode *myco;
{  FreeList((List *) myco->extra);
}

             /* Locate may be either for the server or for a device */
void Port_Locate(myco)
Conode *myco;
{ WORD temp;
  Port_node *node;
             
  if (!strcmp(IOname, myco->name))    /* is it for the server ? */
   { temp = FormOpenReply(Type_Directory, 0L, -1L, -1L); 
     Request_Return(ReplyOK, open_reply, temp);
   }
  elif ( (node = find_node(myco) ) eq (Port_node *) NULL)
   Request_Return(EC_Error + SS_IOProc + EG_Unknown + EO_File, 0L, 0L);
  else
   { temp = FormOpenReply(Type_File, 0L, -1L, -1L);
     Request_Return(ReplyOK, open_reply, temp);
   }
}

void Port_ObjInfo(myco)
Conode *myco;
{ Port_node *node;
                       /* put info in data vector straightaway, nothing in */
                        /* there that we need any more */
  ObjInfo *Heliosinfo = (ObjInfo *) mcb->Data;

             /* The ObjectInfo information is in the Data vector so it does */
             /* not get swapped automatically by the message passing        */
             /* routines. All integers must be swapped explicitly. On the   */
             /* PC swap() is defined to be a no-op.                         */
  Heliosinfo->DirEntry.Flags  = swap(0L);
  Heliosinfo->DirEntry.Matrix = swap(DefFileMatrix);
  Heliosinfo->Size            = swap(0L);
  Heliosinfo->Account         = swap(0L);
  Heliosinfo->Creation        = swap(Startup_Time);
  Heliosinfo->Access          = swap(Startup_Time);
  Heliosinfo->Modified        = swap(Startup_Time);
             
  if (!strcmp(IOname, myco->name))    /* is it for the server ? */
   { Heliosinfo->DirEntry.Type   = swap(Type_Directory);
     Heliosinfo->DirEntry.Matrix = swap(DefDirMatrix);
     strcpy(Heliosinfo->DirEntry.Name, myco->name);
     Request_Return(ReplyOK, 0L, (word) sizeof(ObjInfo));
   }
  elif ( (node = find_node(myco) ) eq (Port_node *) NULL)
   Request_Return(EC_Error + SS_IOProc + EG_Unknown + EO_File, 0L, 0L);
  else
   { Heliosinfo->DirEntry.Type   = swap(Type_File);
     strcpy(Heliosinfo->DirEntry.Name, node->direntry.Name);
     Request_Return(ReplyOK, 0L, (word) sizeof(ObjInfo));
   }
}

/**
*** Rename may be used to change the default port. For example, the user can
*** use "mv /printers/com1 /printers/default" to set the default printer port
*** to be the first serial port. IOname will contain printers/default, and
*** the rename offset will contain default. To set the default printer, all
*** I need to do is change its id in the port node structure. 
**/
void Port_Rename(myco)
Conode *myco;
{ char *data = mcb->Data;
  int  dest    = (int) mcb->Control[RenameToname_off];
  Port_node *source_node;

  if (!strcmp(myco->name, IOname))     /* Cannot rename e.g. /printers */
   { Request_Return(EC_Error + SS_IOProc + EG_WrongFn + EO_Server, 0L, 0L);
     return;
   }

  source_node = find_node(myco);
  if (source_node eq (Port_node *) NULL)
   { Request_Return(EC_Error + SS_IOProc + EG_Unknown + EO_File, 0L, 0L);
     return;
   }

  if (strcmp(&(data[dest]), "default"))
   { Request_Return(EC_Error + SS_IOProc + EG_Protected + EO_File, 0L, 0L);
     return;
   } 

  ((Port_node *) ((List *) myco->extra)->head)->id = source_node->id;
  ((Port_node *) ((List *) myco->extra)->head)->attr = source_node->attr;

  Request_Return(ReplyOK, 0L, 0L);
}

void Port_Open(myco)
Conode *myco;
{ Port_node *node;

  if (!strcmp(IOname, myco->name))    /* is it for the server ? */
   NewStream(Type_Directory, Flags_Closeable, (WORD) myco->extra,
             PortDir_Handlers);
  elif ( (node = find_node(myco) ) eq (Port_node *) NULL)
   Request_Return(EC_Error + SS_IOProc + EG_Unknown + EO_File, 0L, 0L);
  else
   NewStream(Type_File, Flags_Closeable + Flags_Interactive, (WORD) node,
             node->handlers );
}

void Port_TidyStream(myco)
Conode *myco;
{ Port_node *node = (Port_node *) myco->extra;
  (*node->done_fn)(node->id);
  use(myco)
}

void Port_Read(myco)
Conode *myco;
{ Port_node *node = (Port_node *) myco->extra;

  General_PortIO(Port_read, node->id, myco, node->receive_fn, 
                 node->pollread_fn, node->abortread_fn, node->error_fn );
}

void Port_Write(myco)
Conode *myco;
{ Port_node *node = (Port_node *) myco->extra;

  General_PortIO(Port_write, node->id, myco, node->send_fn,
                 node->pollwrite_fn, node->abortwrite_fn, node->error_fn );
}

void Port_Close(myco)
Conode *myco;
{ Port_node *node = (Port_node *) myco->extra;
  (*node->done_fn)(node->id);
  if (mcb->MsgHdr.Reply ne 0L)
    Request_Return(ReplyOK, 0L, 0L);
  Seppuku();
}

void Port_GetAttr(myco)
Conode *myco;
{ Port_node *node = (Port_node *) myco->extra;
  CopyAttributes((Attributes *) mcb->Data, node->attr);
  Request_Return(ReplyOK, 0L, (WORD) sizeof(Attributes));
}

void Port_SetAttr(myco)
Conode *myco;
{ Port_node *node = (Port_node *) myco->extra;
  CopyAttributes(node->attr, (Attributes *) mcb->Data);
  (*node->configure_fn)(node->id, node->attr );
  Request_Return(ReplyOK, 0L, 0L);
}

PRIVATE Port_node *find_node(myco)
Conode *myco;
{ List *list = (List *) myco->extra;
  Port_node *node;
  char *name = &(IOname[0]);

  for ( ; (*name ne '/') && (*name ne '\0'); name++);
  if (*name eq '\0')
    return((Port_node *) NULL);

  *name = '\0';           /* check the server name */
  if (strcmp(IOname, myco->name))
    return((Port_node *) NULL);

  *name++ = '/';          /* server fine, restore IOname */
  for (node = (Port_node *) list->head; node->node.next ne (Node *) NULL;
       node = (Port_node *) node->node.next)
   if (!strcmp(name, node->direntry.Name))
     return(node);

   return((Port_node *) NULL);
}

  
/**
*** A general purpose Port IO function, for use with both reads and writes
*** Note that the coroutine needs its own buffer, because other messages
*** may arrive whilst polling corrupting the mcb.
**/

PRIVATE void General_PortIO(type, ident, myco, start_fn, poll_fn, abort_fn,
                            error_fn)
WORD   type, ident;
Conode *myco;
WordFnPtr start_fn, poll_fn, abort_fn;
VoidFnPtr error_fn;
{ WORD amount, timeout, timelimit, temp;
  BYTE local_buf[80], *buff;
  int  own_buf = 0;

                       /* The offsets are the same for reads and writes */
  amount    = (mcb->Control)[ReadSize_off];     /* how many characters ? */
  timeout   = (mcb->Control)[ReadTimeout_off];
  timelimit = Now + divlong( timeout, time_unit);

                             /* I want a sensible limit on the amount read */
  if ((type eq Port_read) && (amount > IOCDataMax))
    amount = IOCDataMax;

  if (amount eq 0L)           /* It can happen... */
   { mcb->Control[Reply1_off] = 0L;
     Request_Return( (type eq Port_read) ? ReadRc_EOD : ReplyOK,
                     (type eq Port_read) ? 0L : 1L, 0L);
     return;
   }
  if (amount > 0x7FFFL)
   { Request_Return( EC_Error + SS_IOProc + EG_WrongSize + EO_File, 0L, 0L);
     return;
   }

  if (amount <= 80L)
   buff = &(local_buf[0]);
  else
   { buff = (BYTE *) malloc((uint) amount);
     if (buff eq (BYTE *) NULL)
      { Request_Return(EC_Error + SS_IOProc + EG_NoMemory + EO_Server, 0L, 0L);
        return;
      }
     own_buf++;
   }

  if (type eq Port_write)                       /* support multiple writes */
   {
     if (mcb->MsgHdr.DataSize eq 0)
      { WORD fetched   = 0L;
      	BYTE *a_ptr;
        WORD reply_port = mcb->MsgHdr.Reply;

            /* The data is read in in lumps of upto MaxData, */
            /* and put into the same buffer. Once the buffer */
            /* has been filled I start printing it.          */

        mcb->Control[Reply1_off] = (amount > maxdata) ? maxdata : amount;
        mcb->Control[Reply2_off] = maxdata;
        mcb->MsgHdr.Flags = MsgHdr_Flags_preserve; /* must preserve the route here */
        Request_Return(WriteRc_Sizes, 2L, 0L);

        a_ptr = buff;

        while ((fetched < amount) && (timelimit > Now))
         { myco->timelimit = timelimit;
           Suspend();                  /* Wait for the data to be sent */
           if ((myco->type eq CoSuicide) || (myco->type eq CoTimeout))
            { if (own_buf) free(buff);
              Seppuku();
            }
           fetched   += (word) mcb->MsgHdr.DataSize;
           memcpy(a_ptr, mcb->Data, mcb->MsgHdr.DataSize);
           a_ptr     += mcb->MsgHdr.DataSize;           /* next bit of buffer */
           timelimit = Now + divlong( timeout, time_unit);     /* reset timer */
         }

                 /* Now either I have got all the data or I have timed out  */
                 /* Either way I want to print some of it, to prevent       */
                 /* infinite loops of Writing 0 bytes within timeout.       */
                 /* I want to display FETCHED chars */
        amount = fetched;
        mcb->MsgHdr.Reply = reply_port;
      }
     else
       memcpy(buff, &((mcb->Data)[0]), (int) amount);			/* single write */
   }

  unless((*start_fn)(ident, amount, &(buff[0])))
   { if (own_buf) free(buff);
     Request_Return(EC_Recover + SS_IOProc + EG_InUse + EO_Stream, 0L, 0L);
     return;
   }

  AddTail(Remove(&(myco->node)), PollingCo);
                                       /* move from waiting to polling list */
                                       /* because dealing with a request */   
#if multi_tasking
  AddMultiwait(myco, Multi_PortIO, ident);
#else
  myco->type = CoReady;
#endif
  forever
   { if ((temp = (*poll_fn)(ident)) eq -1L)
       { mcb->Control[Reply1_off] = amount;
         break;
       }

     if (temp >= 0L)                      /* An error occurred */
       { (*error_fn)();
         mcb->Control[Reply1_off] = temp;
         if (own_buf) free(buff);
         Request_Return(EC_Error + SS_IOProc + EG_Broken + EO_Stream, 1L, 0L);
         PostInsert(Remove(&(myco->node)), Heliosnode);
                                               /* request */
#if multi_tasking
         ClearMultiwait(myco, Multi_PortIO, ident);
#endif
         return;
       } 

     temp = mcb->MsgHdr.Reply;
     myco->timelimit = timelimit;     
     Suspend();         /* no, and wait by polling */
     mcb->MsgHdr.Reply = temp;

     if (myco->type eq CoSuicide)  /* supposed to die ? */
         { 
           (void) (*abort_fn)(ident);
           if (own_buf) free(buff);
#if multi_tasking
           ClearMultiwait(myco, Multi_PortIO, ident);
#endif
           Seppuku();
         }           
     elif (myco->type eq CoTimeout)    /* time limit exceeded ? */
         { 
           mcb->Control[Reply1_off] = (*abort_fn)(ident);
           break;
         } 
   }

  mcb->MsgHdr.Dest = NullPort;

                      /* The reply message depends on the type of transaction */
  if (type eq Port_read)
   { memcpy(&((mcb->Data)[0]), buff, (int) mcb->Control[Reply1_off]);
     Request_Return(ReadRc_EOD, 0L, mcb->Control[Reply1_off]);
   }
  else
    Request_Return(mcb->Control[Reply1_off] eq amount ? ReplyOK :
                   (EC_Error + SS_IOProc + EG_Timeout + EO_Stream), 1L, 0L);

  if (own_buf) free(buff);

  PostInsert(Remove(&(myco->node)), Heliosnode);
             
#if multi_tasking
  ClearMultiwait(myco, Multi_PortIO, ident);
#endif
}

#endif /* Ports_used */

/*------------------------------------------------------------------------
--
-- the RS232
--
------------------------------------------------------------------------*/

/**
*** The purpose of the Helios RS232 device is to provide just enough support
*** for kermit and for use with serial printers.
**/
#if RS232_supported

/**
*** Time to describe the device-specific functions :
***
*** WORD x_initlist(List *) - initialise device x by fair means or foul, and put
*** Port_node entries into the linked list supplied so that I can provide
*** directory listings of the device etc. This is done most conveniently by
*** calls to add_port_node() above. The routine should return the number of
*** nodes added to the list.
***
*** WORD x_send(WORD id, WORD amount, char *buffer) - start sending the data
*** using the device specified by the id. Each device should be given a unique
*** id in the initlist() routine, and there is also the special case of id 0
*** which is the default device. x_send() is responsible for locking out
*** multiple accesses to the device - if there is an attempt to write to a
*** device that is already being written to then x_send() should return FALSE.
*** Also, if it is not possible to start the IO for any other reason it should
*** return FALSE. If the IO can be started x_send() should return TRUE.
***
*** WORD x_pollwrite(id) - once the IO has been started successfully the device
*** will be polled repeatedly (each time the polling coroutine gets woken up)
*** by a call to x_pollwrite() which includes the id of the device. The routine
*** should return -1L to indicate that the IO has completed successfully, some
*** other negative number to indicate that the IO is proceeding merrily, or
*** a positive number or 0L to indicate an error. This positive number should
*** be the amount of data which you think you managed to write successfully.
*** When an error occurs x_errno should be set to a suitable value, which will
*** be examined by x_error_handler() which will take recovery action for the
*** Helios side - you must do the recovery for the device side.
***
*** WORD x_abortwrite(id) - if the IO needs to be aborted, e.g. because of a
*** timeout, then this function will be called. It should return the amount of
*** data successfully sent.
***
*** WORD x_receive(id, amount, buffer), WORD x_pollread(id), and
*** WORD x_abortread(id) - these are the equivalent routines to x_send() etc.
*** for reading rather than writing. They need only be supplied if the device
*** can be read from as well as written to. For example, a printer is assumed
*** to be always write-only, and a Centronics port may be read from only if
*** specified in the defines.h file.
***
*** void x_done(id) - this is called when the stream to the device is closed,
*** either explicitly or implicitly. You can do whatever you want to in this
*** function.
***
*** void x_configure(id) - set the device to use the
*** attributes specified. These are Helios attributes, so you need to decode
*** them yourself, convert them to whatever local attributes you can implement
*** sensibly, and set these local attributes.
***
*** And if you can write the above nine routines for your hardware then you
*** can support the Helios device. Easy. Just in case you want to look at an
*** example, the devices have been implemented for the PC and for the ST, so
*** see ST/STlocal.c and IBM/PClocal.c .
**/

DirHeader   RS232_list;

void RS232_error_handler()
{ 
}

void RS232_InitServer(myco)
Conode *myco;
{ Port_node *node;

  InitList(&RS232_list.list);

  add_port_node(&RS232_list.list, "default", 0L, NULL);

  RS232_list.entries = RS232_initlist(&RS232_list.list,
     &(((Port_node *) RS232_list.list.head)->attr) ) + 1L;

  myco->extra = (ptr) &RS232_list;

  for (node = (Port_node *) RS232_list.list.head;
       node->node.next ne (Port_node *) NULL;
       node = (Port_node *) node->node.next)
   { Attributes *attr = node->attr;
     InitAttributes(attr);
/**
*** Default attributes : 9600 baud on input and output, 8 bits per byte, with
*** one stop bit, parity disabled on output and ignored on input, xon/xoff
*** enabled on input and output, do not hang up on close, ignore modem status
*** lines, generate event on break.
**/
     SetInputSpeed( attr, (WORD) RS232_B9600);
     SetOutputSpeed(attr, (WORD) RS232_B9600);
     AddAttribute(attr,   RS232_Csize_8);
     AddAttribute(attr,   RS232_IgnPar);
     AddAttribute(attr,   RS232_IXON);
     AddAttribute(attr,   RS232_IXOFF);
     AddAttribute(attr,   RS232_CLocal);
     AddAttribute(attr,   RS232_BreakInterrupt);

     node->send_fn       = func(RS232_send);
     node->pollwrite_fn  = func(RS232_pollwrite);
     node->abortwrite_fn = func(RS232_abortwrite);
     node->receive_fn    = func(RS232_receive);
     node->pollread_fn   = func(RS232_pollread);
     node->abortread_fn  = func(RS232_abortread);
     node->configure_fn  = func(RS232_configure);
     node->done_fn       = func(RS232_done);
     node->error_fn      = func(RS232_error_handler);
     node->handlers      = RS232_Handlers;

     RS232_configure(node->id);
   }
}

void RS232_TidyServer(myco)
Conode *myco;
{ 
#if (PC)
  PUBLIC void fn( RS232_disable_interrupts, (void));
  RS232_disable_interrupts();
#endif
#if (ST)
  extern void RS232_tidy();
  RS232_tidy();
#endif

  Port_TidyServer(myco);
}

void RS232_EnableEvents(myco)
Conode *myco;
{ word mask = mcb->Control[EnableEventsMask_off] &
             (Event_SerialBreak + Event_ModemRing);
  Port_node *node = (Port_node *) myco->extra;

  if (mask eq 0L)       /* disable a break handler */
    { RS232_disable_events(node->id);
      mcb->Control[Reply1_off] = 0L;
#if multi_tasking
      ClearMultiwait(myco, Multi_RS232Event);
#endif
      Request_Return(ReplyOK, 1L, 0L);
    }
  else                  /* install an event handler */
    {
      RS232_enable_events(node->id, mask, mcb->MsgHdr.Reply);
      mcb->Control[Reply1_off] = mask;
      mcb->MsgHdr.Flags = MsgHdr_Flags_preserve;
      Request_Return(ReplyOK, 1L, 0L);
#if multi_tasking
      AddMultiwait(myco, Multi_RS232Event);
#endif
    } 
}


#endif   /* RS232_supported */

/*------------------------------------------------------------------------
--
-- the Centronics interface
--
------------------------------------------------------------------------*/

#if Centronics_supported

DirHeader   Centronics_list;

void Centronics_error_handler()
{ if (Centronics_errno & 0x0010)    /* Check the bit reserved for Centronics errors */
   { switch (Centronics_errno)
      { case printer_outofpaper : output("*** SERVER : printer out of paper.\r\n");
                                  break;
        case printer_offline    : output("*** SERVER : printer off line.\r\n");
                                  break;
        case printer_invalid    : output("*** SERVER : printer not connected.\r\n");
                                  break;

        case printer_error      :
        default                 : output("*** SERVER : unknown error when accessing printer.\r\n");
                                  break;
      }
   }
        
  Centronics_errno = 0;
}

void Centronics_InitServer(myco)
Conode *myco;
{ Port_node *node;
  InitList(&Centronics_list.list);

  add_port_node(&Centronics_list.list, "default", 0L, NULL);

  Centronics_list.entries = Centronics_initlist(&Centronics_list.list,
     &(((Port_node *) Centronics_list.list.head)->attr) ) + 1L;

  myco->extra = (ptr) &Centronics_list;

  for (node = (Port_node *) Centronics_list.list.head;
       node->node.next ne (Port_node *) NULL;
       node = (Port_node *) node->node.next)
   {                                  /* No Centronics attributes are defined */
     InitAttributes(node->attr);
     node->send_fn       = func(Centronics_send);
     node->pollwrite_fn  = func(Centronics_pollwrite);
     node->abortwrite_fn = func(Centronics_abortwrite);
#if Centronics_readable
     node->receive_fn    = func(Centronics_receive);
     node->pollread_fn   = func(Centronics_pollread);
     node->abortread_fn  = func(Centronics_abortread);
#endif
     node->configure_fn  = func(Centronics_configure);
     node->done_fn       = func(Centronics_done);
     node->error_fn      = func(Centronics_error_handler);
     node->handlers      = Centronics_Handlers;
   }
}

#endif /* Centronics_supported */

/*------------------------------------------------------------------------
--
-- the Printer device
--
------------------------------------------------------------------------*/

#if Printer_supported

DirHeader   Printer_list;

void Printer_error_handler()
{ if (Printer_errno & 0x0010)    /* Check the bit reserved for Centronics errors */
   { switch (Printer_errno)
      { case printer_outofpaper : output("*** SERVER : printer out of paper.\r\n");
                                  break;
        case printer_offline    : output("*** SERVER : printer off line.\r\n");
                                  break;
        case printer_invalid    : output("*** SERVER : printer not connected.\r\n");
                                  break;

        case printer_error      :
        default                 : output("*** SERVER : unknown error when accessing printer.\r\n");
                                  break;
      }
   }
        
  Printer_errno = 0;
}

void Printer_InitServer(myco)
Conode *myco;
{ Port_node *node;
  InitList(&Printer_list.list);

  add_port_node(&Printer_list.list, "default", 0L, NULL);

  Printer_list.entries = Printer_initlist(&Printer_list.list,
     &(((Port_node *) Printer_list.list.head)->attr) ) + 1L;

  myco->extra = (ptr) &Printer_list;

  for (node = (Port_node *) Printer_list.list.head;
       node->node.next ne (Port_node *) NULL;
       node = (Port_node *) node->node.next)
   { /* Do not configure the printer ports : it is done already ! */
     node->send_fn       = func(Printer_send);
     node->pollwrite_fn  = func(Printer_pollwrite);
     node->abortwrite_fn = func(Printer_abortwrite);
     node->configure_fn  = func(Printer_configure);
     node->done_fn       = func(Printer_done);
     node->error_fn      = func(Printer_error_handler);
     node->handlers      = Printer_Handlers;
   }
}

#endif /* Printer_supported */

/*------------------------------------------------------------------------
--
-- Midi
--
------------------------------------------------------------------------*/

#if Midi_supported

DirHeader   Midi_list;

void Midi_error_handler()
{ output("in MIDI_error_handler()\r\n");
  Midi_errno = 0;
}

void Midi_InitServer(myco)
Conode *myco;
{ Port_node *node;
  InitList(&Midi_list.list);

  add_port_node(&Midi_list.list, "default", 0L, NULL);

  Midi_list.entries = Midi_initlist(&Midi_list.list) + 1L;;

  myco->extra = (ptr) &Midi_list;

  for (node = (Port_node *) Midi_list.list.head;
       node->node.next ne (Port_node *) NULL;
       node = node->node.next)
   { InitAttributes(&(node->attr));
     Midi_configure(node->id, &(node->attr));
     node->send_fn       = func(Midi_send);
     node->pollwrite_fn  = func(Midi_pollwrite);
     node->abortwrite_fn = func(Midi_abortwrite);
     node->receive_fn    = func(Midi_receive);
     node->pollread_fn   = func(Midi_pollread);
     node->abortread_fn  = func(Midi_abortread);
     node->configure_fn  = func(Midi_configure);
     node->done_fn       = func(Midi_done);
     node->error_fn      = func(Midi_error_handler);
     node->handlers      = Midi_Handlers;
   }
}

#endif    /* Midi_supported */

/*------------------------------------------------------------------------
-- the Mouse and keyboard
------------------------------------------------------------------------*/

/**
*** The mouse device is fairly simple, because all you can do is Open it,
*** enable mouse events, deal with acknowledgements and negative
*** acknowledgements for the mouse, and close it. Open, Close, InitStream,
*** TidyStream, and EnableEvents are all very similar to the corresponding
*** Console routines.
**/

#if mouse_supported

#define       Mouse_limit 256
IOEvent       Mouse_Table[Mouse_limit];
int           Mouse_head, Mouse_tail;
word          Mouse_count, Mouse_ack;
event_handler mouse_handler;

void Mouse_InitServer(myco)
Conode *myco;
{ Mouse_head  = 0;
  Mouse_tail  = 0;
  Mouse_count = 0L;
  Mouse_ack   = -1L;
  mouse_handler.port = (WORD) NULL;
  initialise_mouse();
  use(myco)
}

void Mouse_TidyServer(myco)
Conode *myco;
{ tidy_mouse();
  use(myco)
}

/**
*** The mouse may be optional for some implementations.
**/
#if !(Mouse_Always_Available)
void Mouse_Testfun(result)
bool *result;
{ char *config = get_config("Xsupport");

  if (config ne (char *) NULL)
    *result = 1L;
  else
    *result = 0L;
}
#endif /* !Mouse_Always_Available */

/**
*** new_mouse() is called each time the mouse interrupt routine wishes to add
*** an event to the mouse buffer. It is called from an interrupt routine,
*** so I cannot do any library calls.
**/

void new_mouse(mouse_x, mouse_y, mouse_buttons)
SHORT mouse_x, mouse_y;
WORD  mouse_buttons;
{ IOEvent *curr_mouse;

    /* to avoid ridiculous numbers of mouse events I overwrite previous */
    /* movements. However, I must be careful with the buttons. */

  if ((Mouse_head ne Mouse_tail) &&
      (mouse_buttons eq Buttons_Unchanged) &&
      (Mouse_Table[Mouse_head + (Mouse_limit-1) & 
       (Mouse_limit-1)].Device.Mouse.Buttons eq Buttons_Unchanged))
  { Mouse_head = Mouse_head + (Mouse_limit - 1) & (Mouse_limit - 1);
    Mouse_count--;       /* to keep the counter right */
  }

  curr_mouse = &(Mouse_Table[Mouse_head]);

  curr_mouse->Type                 = Event_Mouse;
  curr_mouse->Counter              = Mouse_count++;
  curr_mouse->Stamp                = Now;
  curr_mouse->Device.Mouse.X       = mouse_x;
  curr_mouse->Device.Mouse.Y       = mouse_y;
  curr_mouse->Device.Mouse.Buttons = mouse_buttons;

  Mouse_head = (Mouse_head + 1) & (Mouse_limit - 1);
  if (Mouse_head eq Mouse_tail)       /* buffer overflow */
    { Mouse_tail = (Mouse_tail + 1) & (Mouse_limit - 1);
      curr_mouse->Type |= Flag_Buffer;
    }
}  

PRIVATE void fn( send_mouse_events, (void));
PRIVATE void poll_the_mouse()
{         /* Are there any mouse events outstanding, and is there a handler */
          /* for the mouse events (i.e. has there been an enable_events) ?  */
  if ((mouse_handler.port ne (WORD) NULL) && (Mouse_tail ne Mouse_head))
    send_mouse_events();
}


/**
*** When poll_the_devices() detects that there are outstanding mouse events
*** and there is a mouse event handler active it calls this routine to send
*** the data off to the event handler on the transputer. There is a limit
*** of IOCDataMax on the amount I can send, so that limits the number of
*** events in the table which I can use up. As usual for event handlers I
*** have to worry about the need for acknowledgements and some data may
*** have got lost due to buffer overflow. The data is copied into a message
*** data area, the message is initialised and sent off, and that is it.
**/

PRIVATE void send_mouse_events()
{ IOEvent *table = (IOEvent *) mcb->Data;
  int     count  = 0;
  word    FnRc   = 0L;

               /* Should the handler acknowledge ? */
  if ((Mouse_count - Mouse_ack) > (Mouse_limit / 2))
    FnRc = EventRc_Acknowledge;

                       /* Copy the events into the message data area,       */
                       /* and check for lost events due to buffer overflow. */
  while ((Mouse_tail ne Mouse_head) && (count < (IOCDataMax / sizeof(IOEvent))))
    { if ((Mouse_Table[Mouse_tail].Type & Flag_Buffer) ne 0L)
        { FnRc |= EventRc_IgnoreLost;
          Mouse_Table[Mouse_tail].Type &= ~Flag_Buffer;
        } 
      copy_event(&(table[count++]), &Mouse_Table[Mouse_tail]);
      Mouse_tail = (Mouse_tail + 1) & (Mouse_limit - 1);
    }

  mcb->MsgHdr.Reply = mouse_handler.port;
  mcb->MsgHdr.Dest  = 0L;
  mcb->MsgHdr.Flags = MsgHdr_Flags_preserve + MsgHdr_Flags_sacrifice;
                                                             /* and send it */
  Request_Return(FnRc, 0L, (word)(sizeof(IOEvent) * count));
}

/**
*** opening a mouse is very straightforward.
**/

void Mouse_Open(myco)
Conode *myco;
{ if (strcmp(IOname, "mouse") )
    { Request_Return(EC_Error + SS_IOProc + EG_Name + EO_Message, 0L, 0L);
      return;
    }

  NewStream(Type_File, Flags_Closeable + Flags_Interactive, NULL, 
            Mouse_Handlers);
  use(myco)
}

WORD Mouse_InitStream(myco)
Conode *myco;
{ myco->extra = (WORD *) NULL;       /* A new stream does not own the event handler */
  return(ReplyOK);
}

WORD Mouse_TidyStream(myco)
Conode *myco;
{          /* does this stream own mouse events ?*/
           /* The stream must not time out if it still owns the event handler */
  if ((event_handler *) myco->extra ne (event_handler *) NULL)
   return(1L);
  else
   return(0L);
}

void Mouse_Close(myco)
Conode *myco;
{ if (mcb->MsgHdr.Reply ne 0L)
    Request_Return(ReplyOK, 0L, 0L);
  if ((event_handler *) myco->extra ne (event_handler *) NULL)
   { ((event_handler *) myco->extra)->port = 0L;
     stop_mouse();
   }
  Seppuku();
}

void Mouse_EnableEvents(myco)
Conode *myco;
{ word mask = mcb->Control[EnableEventsMask_off] & Event_Mouse;

  if (mask eq 0L)                             /* disable a mouse handler */
    { if (mouse_handler.port ne (word) NULL)  /* is there a mouse handler ? */
       { mouse_handler.port = (word) NULL;
         *mouse_handler.ownedby = (word) NULL;
         stop_mouse();
#if multi_tasking
         ClearMultiwait(myco, Multi_MouseInput);
#endif
       }
      mcb->Control[Reply1_off] = 0L;
      Request_Return(ReplyOK, 1L, 0L);
    }
  else                                       /* install an event handler */
    { if (mouse_handler.port ne (word) NULL) /* disable an existing handler */
         *mouse_handler.ownedby = (word) NULL;
      else
        start_mouse();
      mouse_handler.port = mcb->MsgHdr.Reply;
      mouse_handler.ownedby = (WORD *) &(myco->extra);
      ((event_handler *)myco->extra)  = &mouse_handler;
      mcb->Control[Reply1_off] = mask;
      mcb->MsgHdr.Flags = MsgHdr_Flags_preserve;
#if multi_tasking
      AddMultiwait(myco, Multi_MouseInput);
#endif
      Request_Return(ReplyOK, 1L, 0L);
    } 
}

/**
*** Mouse_acknowledge is used to tell me that I can release some of my buffer
*** space. Because my buffers are statically allocated and I allow for buffer
*** overflow anyway it is not essential. However, I suspect that its use
*** reduces the frequency of the Ignore_Lost flag and hence the system will
*** recover better if a message is lost. I do not check the validity of the
*** number acknowledged because this allows the other end to acknowledge
*** upto MaxInt events immediately.
**/

void Mouse_Acknowledge(myco)
Conode *myco;
{ Mouse_ack  = mcb->Control[AcknowledgeCount_off];
  use(myco)
}

/**
*** Mouse_NegAcknowledge is used to indicate that one or more messages has been
*** lost. The mcb provided contains the number of the next message expected,
*** and I have to reset my tail pointer to correspond to that event. Then
*** poll_the_devices() will take care of sending the data again.
***
*** There are severe problems if the buffer is almost full, the negative
*** acknowledgement takes me back to an event that has been overwritten or is
*** about to be overwritten, and a new mouse interrupt occurs... I need
*** semaphore protection and truly parallel processes to do this properly, but
*** I have to make do with an ST or a PC. Hence I can but hope that this routine
*** will never cause a crash, or not regularly enough to trace the problem to
*** me....
**/

void Mouse_NegAcknowledge(myco)
Conode *myco;
{ word count    = mcb->Control[NegAcknowledgeCount_off];
  word curcount = Mouse_Table[Mouse_tail].Counter;

                                   /* validate the event count */
  if (count > curcount) return;    /* cannot negacknowledge an unsent message */

/**
*** Now move backwards through Mouse_table until I reach the current head
*** pointer or until the current event is the one after the one acknowledged.
**/
  while (curcount ne count)
   { if (Mouse_tail eq ((Mouse_head + 1) & (Mouse_limit - 1)))
       break;      /* cannot go back any further */
           /* Yukk, I need semaphores here but I have not got them */
     Mouse_tail = (Mouse_tail - 1) & (Mouse_limit - 1);
     curcount   = Mouse_Table[Mouse_tail].Counter;
   }

  if (curcount ne count)                  /* Not reached correct event */
    Mouse_Table[Mouse_tail].Type |= Flag_Buffer;   /* so buffer overflow */

                /* A negative acknowledgement tells me which message has been */
                /* received. */
  Mouse_ack = count - 1;

  use(myco)
}

#endif       /* mouse_supported */

/*-----------------------------------------------------------------------
--
-- The raw keyboard device
--
-----------------------------------------------------------------------*/

#if keyboard_supported
/**
*** There is no real difference between this device and the above mouse
*** device.
**/

#define       Keyboard_limit 256
IOEvent       Keyboard_Table[Keyboard_limit];
int           Keyboard_head, Keyboard_tail;
word          Keyboard_count, Keyboard_ack;
event_handler keyboard_handler = { 0L, (WORD *) NULL };

PRIVATE WORD Repeat_interval;        /* time interval for auto-repeat */
PRIVATE WORD Repeat_delay;           /* delay before auto-repeat starts */
PRIVATE void fn ( send_keyboard_events, (void));

void Keyboard_InitServer(myco)
Conode *myco;
{ Repeat_interval = get_int_config("repeat_interval");
  if (Repeat_interval eq Invalid_config) 
    Repeat_interval = 15L;
  
  Repeat_delay = get_int_config("repeat_delay");
  if (Repeat_delay eq Invalid_config)
    Repeat_delay = 50L;

  Keyboard_head  = 0;
  Keyboard_tail  = 0;
  Keyboard_count = 0L;
  Keyboard_ack   = -1L;
  keyboard_handler.port = (WORD) NULL;

  initialise_keyboard();
  use(myco)
}

void Keyboard_TidyServer(myco)
Conode *myco;
{ tidy_keyboard();
  keyboard_handler.port = (WORD) NULL;
  use(myco)
}
 
#if !(Keyboard_Always_Available)
void Keyboard_Testfun(result)
bool *result;
{ char *config = get_config("Xsupport");

  if (config ne (char *) NULL)
    *result = 1L;
  else
    *result = 0L;
}
#endif /* !Keyboard_Always_Available */

/**
*** new_keyboard() is called to add a new keyboard event to the buffer. The
*** information required is which key and whether the event was key-up or
*** key-down.
**/

void new_keyboard(event, key)
WORD event, key;
{ IOEvent *curr_keyboard = &(Keyboard_Table[Keyboard_head]);
  curr_keyboard->Type    = Event_Keyboard;
  curr_keyboard->Counter = Keyboard_count++;
  curr_keyboard->Stamp   = Now;
  curr_keyboard->Device.Keyboard.What = (WORD) event;
  curr_keyboard->Device.Keyboard.Key  = (WORD) key;
  Keyboard_head = (Keyboard_head + 1) & (Keyboard_limit - 1);
  if (Keyboard_head eq Keyboard_tail)
    { Keyboard_tail = (Keyboard_tail + 1) & (Keyboard_limit - 1);
      curr_keyboard->Type |= Flag_Buffer;
    }
}  

/**
*** This routine checks whether or not there is a keyboard handler installed.
*** If so and there are outstanding events, these are sent. Otherwise the
*** routine returns 1 to indicate to poll_the_devices above that it should
*** poll every window for input.
**/
PRIVATE int poll_the_keyboard()
{ if (keyboard_handler.port ne (WORD) NULL)
    { if (Keyboard_tail ne Keyboard_head)
        send_keyboard_events();
      return(0);
    }
  else
   return(1);
}

/**
*** The code for sending keyboard events is rather similar to that for mouse
*** ones because I copied it from above.
**/

PRIVATE void send_keyboard_events()
{ IOEvent *table = (IOEvent *) mcb->Data;
  int     count = 0;
  word    FnRc  = 0L;

  if ((Keyboard_count - Keyboard_ack) > (Keyboard_limit / 2))
    FnRc = EventRc_Acknowledge;

  while ((Keyboard_tail ne Keyboard_head) &&
         (count < (IOCDataMax / sizeof(IOEvent))))
    { if ((Keyboard_Table[Keyboard_tail].Type & Flag_Buffer) ne 0L)
        { FnRc |= EventRc_IgnoreLost;
          Keyboard_Table[Keyboard_tail].Type &= ~Flag_Buffer;
        } 

      if (debugflags & Keyboard_Flag)
       ServerDebug("Sending raw keyboard event : %lx %s",
                Keyboard_Table[Keyboard_tail].Device.Keyboard.Key,
                ((Keyboard_Table[Keyboard_tail].Device.Keyboard.What eq
                  Keys_KeyUp) ? "Up" : "Down") );

      copy_event(&(table[count++]), &Keyboard_Table[Keyboard_tail]);
      Keyboard_tail = (Keyboard_tail + 1) & (Keyboard_limit - 1);

    }

  mcb->MsgHdr.Reply = keyboard_handler.port;
  mcb->MsgHdr.Dest  = 0L;
  mcb->MsgHdr.Flags = MsgHdr_Flags_preserve;

  Request_Return(FnRc, 0L, (word)(sizeof(IOEvent) * count));
}

/**
*** Here are the GSP handlers for the keyboard.
**/

void Keyboard_Open(myco)
Conode *myco;
{ if (strcmp(IOname, "keyboard") )
    { Request_Return(EC_Error + SS_IOProc + EG_Name + EO_Message, 0L, 0L);
      return;
    }

  NewStream(Type_File, Flags_Closeable + Flags_Interactive, NULL, 
            Keyboard_Handlers);
  use(myco)
}

WORD Keyboard_InitStream(myco)
Conode *myco;
{ myco->extra = (WORD *) NULL;  /* stream does not own keyboard event handler */
  return(ReplyOK);
}

WORD Keyboard_TidyStream(myco)
Conode *myco;
{                                /* does this stream own keyboard events ?*/
  if ((event_handler *)myco->extra ne (event_handler *) NULL)
    return(1L);
  else
    return(0L);
}

void Keyboard_Close(myco)
Conode *myco;
{ if (mcb->MsgHdr.Reply ne 0L)
    Request_Return(ReplyOK, 0L, 0L);
  if ((event_handler *) myco->extra ne (event_handler *) NULL)
   { stop_keyboard();
     ((event_handler *) myco->extra)->port = 0L;
   }
  Seppuku();
}

void Keyboard_EnableEvents(myco)
Conode *myco;
{ word mask = mcb->Control[EnableEventsMask_off] & Event_Keyboard;

  if (mask eq 0L)       /* disable a keyboard handler */
    { if (keyboard_handler.port ne (word)NULL)  /* is there a keyboard handler ? */
       { stop_keyboard();
         keyboard_handler.port = (word) NULL;
         *keyboard_handler.ownedby = (word) NULL;
#if multi_tasking
         ClearMultiwait(myco, Multi_KeyboardInput);
#endif
       }
      mcb->Control[Reply1_off] = 0L;
      Request_Return(ReplyOK, 1L, 0L);
    }
  else                 /* install an event handler */
    { if (keyboard_handler.port ne (word) NULL) /* disable an existing handler */
        *keyboard_handler.ownedby = (word) NULL;
      else
       start_keyboard();
      keyboard_handler.port = mcb->MsgHdr.Reply;
      keyboard_handler.ownedby = (WORD *) &(myco->extra);
      ((event_handler *)myco->extra)  = &keyboard_handler;
      mcb->Control[Reply1_off] = mask;
      mcb->MsgHdr.Flags = MsgHdr_Flags_preserve;
#if multi_tasking
      AddMultiwait(myco, Multi_KeyboardInput);
#endif
      Request_Return(ReplyOK, 1L, 0L);
    } 
}

void Keyboard_Acknowledge(myco)
Conode *myco;
{ Keyboard_ack  = mcb->Control[AcknowledgeCount_off];
  use(myco)
}

void Keyboard_NegAcknowledge(myco)
Conode *myco;
{ word count    = mcb->Control[NegAcknowledgeCount_off];
  word curcount = Keyboard_Table[Keyboard_tail].Counter;

                                   /* validate the event count */
  if (count > curcount) return;    /* cannot negacknowledge an unsent message */

/**
*** Now move backwards through Mouse_table until I reach the current head
*** pointer or until the current event is the one after the one acknowledged.
**/
  while (curcount ne count)
   { if (Keyboard_tail eq ((Keyboard_head + 1) & (Keyboard_limit - 1)))
       break;      /* cannot go back any further */
           /* Yukk, I need semaphores here but I have not got them */
     Keyboard_tail = (Keyboard_tail - 1) & (Keyboard_limit - 1);
     curcount   = Keyboard_Table[Keyboard_tail].Counter;
   }

  if (curcount ne count)                         /* Not reached correct event */
    Keyboard_Table[Keyboard_tail].Type |= Flag_Buffer;  /* so buffer overflow */

                /* A negative acknowledgement tells me which message has been */
                /* received. */
  Keyboard_ack = count - 1;
  use(myco)
}

#endif /* keyboard_supported */
