/*------------------------------------------------------------------------
--                                                                      --
--          H E L I O S   I N P U T / O U T P U T   S E R V E R         --
--          ---------------------------------------------------         --
--                                                                      --
--               Copyright (C) 1987, Perihelion Software Ltd.           --
--                             All Rights Reserved.                     --
--                                                                      --
--  terminal.c                                                          --
--                                                                      --
--           This module contains code to interact with the various     --
--                                                                      --
--           forms of terminal devices.                                 --
--                                                                      --
--  Author:  BLV 19/10/88                                               --
--                                                                      --
------------------------------------------------------------------------*/
/* SccsId: 3.7 14/4/89\ Copyright (C) 1988, Perihelion Software Ltd.        */

/**
*** The usual header file and some function declarations.
**/

#define Terminal_Module
#include "helios.h"


/*------------------------------------------------------------------------
--------------------------------------------------------------------------
--
--     The console and window devices
--
--------------------------------------------------------------------------
------------------------------------------------------------------------*/

/**
*** These devices are probably the most complicated one that I need to
*** support. There are lots of nasties like converting ANSI escape sequences,
*** the Helios standard, to local ones, or like the user XOFF-ing at the
*** most inconvenient moments.
***
***  Console attributes. Attributes act globally on the
*** entire console device. If sensible windows are available each
*** window has its own set of attributes. Global attributes are
*** necessary to deal with XON/XOFF, i.e. whenever one Stream changes its
*** attributes this affects all the other streams. The structure Console_attr,
*** declared at the top of this module, is used to hold these global attributes.
**/

/**
*** In the absence of multiple windows, there is a single window structure
*** Server_window which is used to store everything. InitServer() has to
*** initialise this structure completely.
**/
void Console_InitServer(myco)
Conode *myco;
{
  InitAttributes (&(Server_window.attr));
  AddAttribute   (&(Server_window.attr), ConsoleEcho);
  RemoveAttribute(&(Server_window.attr), ConsoleIgnoreBreak);
  AddAttribute   (&(Server_window.attr), ConsoleBreakInterrupt);
  AddAttribute   (&(Server_window.attr), ConsolePause);
  RemoveAttribute(&(Server_window.attr), ConsoleRawInput);
  RemoveAttribute(&(Server_window.attr), ConsoleRawOutput);
  Server_window.break_handler.port    = NullPort;
  Server_window.break_handler.ownedby = (WORD *) NULL;
  Server_window.cooker.count          = 0;
  Server_window.cooker.outstanding    = 0;
  Server_window.cooker.EOFco          = -1L;
  Server_window.cooker.readerco       = (Conode *) NULL;
  Server_window.head                  = 0;
  Server_window.tail                  = 0;
  Server_window.XOFF                  = 0;
  Server_window.stream_count          = 10000;
  use(myco)
}

/**
*** Console_Open is responsible for creating a new Stream coroutine, which is
*** always possible if there is enough memory. All the entries in the coroutine
*** list node are set up, and the reply is set up. Flags_Interactive is a
*** hint to the transputer side that data should be read and written in fairly
*** small lumps.
***
*** At any moment in time there are likely to be lots of open console Streams,
*** because every C program including the shell has at least three for STDIN,
*** STDOUT, and STDERR. Hence all Console streams must refer to the same window.
**/

void Console_Open(myco)
Conode *myco;
{                                 /* check the device we are opening */
  if (strcmp(IOname, "console") )
    { Request_Return(EC_Error + SS_IOProc + EG_Name + EO_Message, 0L, 0L);
      return;
    }

  NewStream(Type_File, Flags_Closeable + Flags_Interactive, (WORD) &Server_window, 
            Console_Handlers);
  use(myco)
}

#if multiple_windows

/**
*** The window device supports multiple console streams on a single
*** IO processor. It depends on a number of local routines, as follows :
***
*** WORD create_a_window(char *name) - this should create a new window
*** somewhere on the screen, with the name specified. This name might appear
*** in the window's title bar, for example, or it might be ignored. It
*** should return a handle of some sort.
***
*** void close_window(handle) - this should destroy the window specified.
*** The handle will be something returned by create_a_window.
***
*** void window_size(handle, word *x, word *y) - this should store the current 
*** window size in the specified variables.
***
*** The Window server routines follow next. They are rather different from
*** the console ones because the window server is a directory, i.e. much
*** more complicated.
**/

/**
*** Test functions used to determine whether or not to support multiple
*** windows.
**/

PRIVATE bool use_windows = -1L;

void Window_Testfun(result)
bool *result;
{ char *temp;

  if (use_windows ne -1L)
   { *result = use_windows; return; }

  temp = get_config("server_windows");
  if (temp ne NULL)
   use_windows = 1L;
  else
   use_windows = 0L;

  *result = use_windows; 
}

void Console_Testfun(result)
bool *result;
{ char *temp;

  if (use_windows ne -1L)
   { *result = !use_windows; return; }

  temp = get_config("server_windows");
  if (temp ne NULL)
   use_windows = 1L;
  else
   use_windows = 0L;

  *result = !use_windows;
}

/**
*** Window_InitServer() sets up the linked list of windows, initially Helios
*** has not created any windows so the list is empty. A separate window is
*** created for the Server's output, see initialise_devices() in devices.c .
**/
PRIVATE int Next_WindowID;

void Window_InitServer(myco)
Conode *myco;
{ InitList(&(Window_List.list));
  Window_List.entries = 0L;
  Next_WindowID = 1;
  use(myco);
}

/**
*** TidyServer goes through all the windows created by Helios and deletes
*** them all.
**/
void Window_TidyServer(myco)
Conode *myco;
{ Window *window;

  for (window = (Window *) Window_List.list.head;
       window->node.node.next ne (Node *) NULL;
       window = (Window *) window->node.node.next )
   { close_window(window->handle);
     free(Remove(&(window->node.node)));
   }
  use(myco)
}

/**
*** new_window tries to create a new window with the name given, fully
*** initialising the window structure. If it fails it sends a suitable
*** error message to the transputer, otherwise it returns TRUE and
*** leaves it to the caller, Window_Open or Window_Create, to send a
*** suitable reply.
**/
PRIVATE Window *new_window(window_name)
char *window_name;
{ Window *window = (Window *) malloc(sizeof(Window));
  if (window eq (Window *) NULL)
   { Request_Return(EC_Warn + SS_IOProc + EG_NoMemory + EO_Server, 0L, 0L);
     return(window);
   }

  window->handle = create_a_window(window_name);

  if (window->handle eq 0L)
   { free(window);
     Request_Return(EC_Warn + SS_IOProc + EG_Create + EO_Window, 0L, 0L);
     return((Window *) NULL);
   }
  strcpy(&(window->node.direntry.Name[0]), window_name);
  window->node.direntry.Type = Type_File;
  window->XOFF   = 0;
  window->head   = 0;
  window->tail   = 0;
  InitAttributes(&(window->attr));
  AddAttribute   (&(window->attr), ConsoleEcho);
  RemoveAttribute(&(window->attr), ConsoleIgnoreBreak);
  AddAttribute   (&(window->attr), ConsoleBreakInterrupt);
  AddAttribute   (&(window->attr), ConsolePause);
  RemoveAttribute(&(window->attr), ConsoleRawInput);
  RemoveAttribute(&(window->attr), ConsoleRawOutput);
  window->break_handler.port    = NullPort;
  window->break_handler.ownedby = 0L;
  window->stream_count          = 0;   
  window->cooker.count          = 0;
  window->cooker.outstanding    = 0;
  window->cooker.EOFco          = -1L;
  window->cooker.readerco       = (Conode *) NULL;
  Window_List.entries++;
#if use_ANSI_emulator
  { word x, y;
    window_size(window->handle, &x, &y);
    if (!Init_Ansi(window, (int) y, (int) x))
     { close_window(window->handle);
       free(window);
       Request_Return(EC_Warn + SS_IOProc + EG_NoMemory + EO_Server, 0L, 0L);
       return((Window *) NULL);
     }
  }
#endif  
  AddTail(&(window->node.node), &(Window_List.list));

#if (use_ANSI_emulator && (!real_windows))
  sprintf(misc_buffer1, "%c0;%2dH%c7m%s%c0m\r\n", 0x009B,
          81 - strlen(window_name), 0x009B, window_name, 0x009B); 
  window_output(misc_buffer1, window);
#endif

  return(window);
}

/**
*** An open request for window means that somebody wants to look at the
*** window directory, i.e. find out what windows currently exist. An open
*** request for an existing window means that somebody wants to read from
*** and write to a real window. An Open request for a non-existant window
*** is satisfied iff the Create bit is set in the open mode, and a new
*** window is created automatically.
**/ 
void Window_Open(myco)
Conode *myco;
{ char   *window_name;
  Window *window;
  WORD   openmode = mcb->Control[OpenMode_off];

                                         /* check the device we are opening */
  if (!strcmp(IOname, "window") )        /* Opening the directory ? */
    { NewStream(Type_Directory, Flags_Closeable, (WORD) &Window_List,  
              WindowDir_Handlers);
      return;
    }

             /* Extract the name of the window from IOname */
  for (window_name = IOname; (*window_name ne '/') && (*window_name ne '\0');
       window_name++);
  if (*window_name eq '\0')
    { Request_Return(EC_Error + SS_IOProc + EG_Name + EO_Message, 0L, 0L);
      return;
    }

  window_name++;
                  /* Try to find the name in the list */
  for (window = (Window *) Window_List.list.head;
       window->node.node.next ne (Node *) NULL;
       window = (Window *) window->node.node.next )
    if (!mystrcmp(window_name, &(window->node.direntry.Name[0])))
     { NewStream(Type_File, Flags_Closeable + Flags_Interactive,
                 (WORD) window, Console_Handlers);
       return;
     }

  if (openmode & O_Create)
   { if ((window = new_window(window_name)) ne (Window *) NULL)
      NewStream(Type_File, Flags_Closeable + Flags_Interactive,
                (WORD) window, Console_Handlers);
   }
  else
   Request_Return(EC_Error + SS_IOProc + EG_Unknown + EO_Window, 0L, 0L);
  use(myco)
}

/**
*** Window_Locate may look for either the window server or for one of the
*** windows in the directory.
**/
void Window_Locate(myco)
Conode *myco;
{ char *window_name;
  Window *window;
  WORD   temp;
         
  if (!strcmp(IOname, "window") )
   { temp = FormOpenReply(Type_Directory, 0L, -1L, -1L); 
     Request_Return(ReplyOK, open_reply, temp);
     return;
   }

             /* Extract the name of the window from IOname */
  for (window_name = IOname; (*window_name ne '/') && (*window_name ne '\0');
       window_name++);
  if (*window_name eq '\0')
    { Request_Return(EC_Error + SS_IOProc + EG_Name + EO_Message, 0L, 0L);
      return;
    }

  window_name++;
                      /* Try to find the name in the list */
  for (window = (Window *) Window_List.list.head;
       window->node.node.next ne (Node *) NULL;
       window = (Window *) window->node.node.next )
    if (!mystrcmp(window_name, &(window->node.direntry.Name[0])))
     { temp = FormOpenReply(Type_File, 0L, -1L, -1L);
       Request_Return(ReplyOK, open_reply, temp);
       return;
     }

  Request_Return(EC_Error + SS_IOProc + EG_Unknown + EO_Window, 0L, 0L);
  use(myco)
}

/**
*** Window_Create may be used to guarantee the existance of a particular
*** window or of the window server, i.e. act just like Locate. Alternatively
*** it may be used to create a new window.
**/
void Window_Create(myco)
Conode *myco;
{ char *window_name;
  Window *window;
  WORD   temp;
         
  if (!strcmp(IOname, "window") )
   { temp = FormOpenReply(Type_Directory, 0L, -1L, -1L); 
     Request_Return(ReplyOK, open_reply, temp);
     return;
   }

             /* Extract the name of the window from IOname */
  for (window_name = IOname; (*window_name ne '/') && (*window_name ne '\0');
       window_name++);
  if (*window_name eq '\0')
    { Request_Return(EC_Error + SS_IOProc + EG_Name + EO_Message, 0L, 0L);
      return;
    }

  window_name++;

                      /* Try to find the name in the list */
  for (window = (Window *) Window_List.list.head;
       window->node.node.next ne (Node *) NULL;
       window = (Window *) window->node.node.next )
    if (!strcmp(window_name, &(window->node.direntry.Name[0])))
     { sprintf(&(misc_buffer1[0]), "%s.%d", IOname, Next_WindowID++);
       strcpy(IOname, &(misc_buffer1[0]));
       break;
     }

  if (new_window(window_name) ne (Window *) NULL)
   { temp = FormOpenReply(Type_File, 0L, -1L, -1L);
     Request_Return(ReplyOK, open_reply, temp);
   }
  use(myco)
}

/**
*** Delete is used to destroy windows. The window server cannot be
*** removed.
**/
void Window_Delete(myco)
Conode *myco;
{ char *window_name;
  Window *window;
         
  if (!strcmp(IOname, "window") )
   { Request_Return(EC_Error + SS_IOProc + EG_InUse + EO_Server, 0L, 0L);
     return;
   }
             /* Extract the name of the window from IOname */
  for (window_name = IOname; (*window_name ne '/') && (*window_name ne '\0');
       window_name++);
  if (*window_name eq '\0')
    { Request_Return(EC_Error + SS_IOProc + EG_Name + EO_Message, 0L, 0L);
      return;
    }

  window_name++;
                    /* Try to find the name in the list */
  for (window = (Window *) Window_List.list.head;
       window->node.node.next ne (Node *) NULL;
       window = (Window *) window->node.node.next )
    if (!mystrcmp(window_name, &(window->node.direntry.Name[0])))
     {
       if (window->stream_count ne 0)
        { Request_Return(EC_Error + SS_IOProc + EG_InUse + EO_Stream, 0L, 0L);
          return; 
        }
       close_window(window->handle);
#if use_ANSI_emulator
       Tidy_Ansi(&(window->screen));
#endif
       free(Remove(&(window->node.node)));
       Window_List.entries--;
       Request_Return(ReplyOK, 0L, 0L);
       return;
     }
  Request_Return(EC_Error + SS_IOProc + EG_Unknown + EO_Window, 0L, 0L);
  use(myco)
}

/**
*** ObjectInfo is slightly different depending on whether it is for the
*** server or for one of the windows.
**/
void Window_ObjectInfo(myco)
Conode *myco;
{ char *window_name;
  Window *window;
  ObjInfo *Heliosinfo = (ObjInfo *) mcb->Data;

             /* The ObjectInfo information is in the Data vector so it does */
             /* not get swapped automatically by the message passing        */
             /* routines. All integers must be swapped explicitly. On the   */
             /* PC swap() is defined to be a no-op.                         */
  Heliosinfo->DirEntry.Type   = swap(Type_File);
  Heliosinfo->DirEntry.Flags  = swap(0L);
  Heliosinfo->DirEntry.Matrix = swap(DefFileMatrix);
  Heliosinfo->Account         = swap(0L);
  Heliosinfo->Creation        = swap(Startup_Time);
  Heliosinfo->Access          = swap(Startup_Time);
  Heliosinfo->Modified        = swap(Startup_Time);
             
  if (!strcmp(IOname, "window") )
   { Heliosinfo->DirEntry.Type   = swap(Type_Directory);
     Heliosinfo->DirEntry.Matrix = swap(DefDirMatrix);
     strcpy(Heliosinfo->DirEntry.Name, "window");
     Heliosinfo->Size            = swap(0L);
     Request_Return(ReplyOK, 0L, (word) sizeof(ObjInfo));
     return;
   }

             /* Extract the name of the window from IOname */
  for (window_name = IOname; (*window_name ne '/') && (*window_name ne '\0');
       window_name++);
  if (*window_name eq '\0')
    { Request_Return(EC_Error + SS_IOProc + EG_Name + EO_Message, 0L, 0L);
      return;
    }

  window_name++;
                   /* Try to find the name in the list */
  for (window = (Window *) Window_List.list.head;
       window->node.node.next ne (Node *) NULL;
       window = (Window *) window->node.node.next )
    if (!mystrcmp(window_name, &(window->node.direntry.Name[0])))
     { strcpy(Heliosinfo->DirEntry.Name, window->node.direntry.Name);
       Heliosinfo->Size            = swap((WORD) window->stream_count);
       Request_Return(ReplyOK, 0L, (word) sizeof(ObjInfo));
       return;
     }

  Request_Return(EC_Error + SS_IOProc + EG_Unknown + EO_Window, 0L, 0L);
  use(myco);
}
#endif

/*------------------------------------------------------------------------
--------------------------------------------------------------------------
--
--     console and window streams
--
--------------------------------------------------------------------------
------------------------------------------------------------------------*/

/**
*** InitStream() is called when a new stream is opened. The extra field
*** in the coroutine list node points to the parent window, or to
*** Server_window if multiple windows are not available.
**/

WORD Console_InitStream(myco)
Conode *myco;
{
#if multiple_windows
  Window *window = (Window *) myco->extra;
  window->stream_count++;
#else
  use(myco)
#endif
 
  return(ReplyOK);
}


/**
*** TidyStream() is called when the Server exits or reboots.
**/

WORD Console_TidyStream(myco)
Conode *myco;
{ Window *window = (Window *) myco->extra;

#if multiple_windows
  window->stream_count--;
#endif

  if (window->break_handler.ownedby eq (WORD *) myco)
   { window->break_handler.port    = NullPort;
     window->break_handler.ownedby = (WORD *) NULL;
   }

  return(0L);
}

/**
*** Closing a console Stream is relatively easy, because there are no
*** file handles that need closing or anything like that. 
**/

void Console_Close(myco)
Conode *myco;
{
  if (mcb->MsgHdr.Reply ne 0L)
    Request_Return(ReplyOK, 0L, 0L);

  (void) Console_TidyStream(myco);

  Seppuku();
}

/**
*** GetAttributes is responsible for sending off the current global attributes,
*** held in the parent window structure. Care has to be taken to get the
*** window sizes right. CopyAttributes()is used because it does all the
*** required swapping.
**/

void Console_GetAttr(myco)
Conode *myco;
{ Window *window = (Window *) myco->extra;

#if multiple_windows
  word x, y;
  window_size(window->handle, &x, &y);
  window->attr.Time = (short) x;
  window->attr.Min  = (short) y;
#else
#if (ST || PC)
  window->attr.Min  = 25;
  window->attr.Time = 80;
#endif
#endif  /* multiple_windows */

  CopyAttributes((Attributes *) mcb->Data, &(window->attr));
  Request_Return(ReplyOK, 0L, (WORD) sizeof(Attributes));
}

/**
*** SetAttributes just takes the new attributes which were part of the message
*** and makes them global. I have to make certain that a window is not left
*** permanently XOFF'ed due to a change of attributes.
**/

void Console_SetAttr(myco)
Conode *myco;
{ Window *window = (Window *) myco->extra;

  CopyAttributes(&(window->attr), (Attributes *) mcb->Data);

  if (!IsAnAttribute(&(window->attr), ConsolePause))
   window->XOFF = 0;

  Request_Return(ReplyOK, 0L, 0L);
}

/**
*** The only event that exists for a console stream is break events. See the
*** chapter on Data and Devices in the technical manual for details of this.
*** There are various things to consider. First of all, the request may be used
*** to disable an existing event handler by masking off the appropriate event,
*** but it is perfectly feasible for the Stream not to have the event handler
*** anymore. If the request does involve setting up an event handler I must
*** release any existing handler, and the structure break_handler contains a
*** pointer to the existing stream which owns the event. Then I set up this
*** stream to own the event handler and the event handler to be owned by this
*** stream. I hope the last sentence made sense. Finally I can reply to the
*** request.
***
*** It is not worth worrying about acknowledgements and negative acknowledge for
*** break events, so they are not supported.
**/

void Console_EnableEvents(myco)
Conode *myco;
{ word mask = mcb->Control[EnableEventsMask_off] & Event_Break;
  Window *window = (Window *) myco->extra;
  event_handler *handler = &(((Window *) myco->extra)->break_handler);

  if (mask eq 0L)       /* disable a break handler */
    { if (handler->port ne (word)NULL)    /* is there a break handler ? */
       { handler->port = (word) NULL;
         handler->ownedby = (word *) NULL;
       }
      mcb->Control[Reply1_off] = 0L;
      Request_Return(ReplyOK, 1L, 0L);
#if multi_tasking
      ClearMultiwait(myco, Multi_ConsoleInput, window->handle);
#endif
    }
  else                  /* install an event handler */
    { handler->port = mcb->MsgHdr.Reply;  /* and install new one */
      handler->ownedby = (WORD *) myco;
      mcb->Control[Reply1_off] = mask;
      mcb->MsgHdr.Flags = MsgHdr_Flags_preserve;
      Request_Return(ReplyOK, 1L, 0L);
#if multi_tasking
      AddMultiwait(myco, Multi_ConsoleInput, window->handle);
#endif
    } 
}

/**
*** Reading from the keyboard is absolutely lovely. It involves extracting
*** data from the circular buffer where it will have been put by
*** poll_the_devices(), processing it if in cooked mode, and sending it off.
*** I have to be very careful with timeouts.
***
*** To extract data from the circular buffer I use a single routine
*** window_getchar(), which is the opposite of add_key(). The routine
*** returns 0 if there is no character ready, else the character, on the
*** reasoning that it is impossible to type '\0' on any normal keyboard and
*** if it were possible the system would crash anyway.
**/

PRIVATE void fn( raw_Console_read,    (Conode *, int, WORD));
PRIVATE void fn( cooked_Console_read, (Conode *, int, WORD));

              /* extract a single character from the console buffer, if any */
int window_getchar(window)
Window *window;
{ int tempch;

  if (window->head eq window->tail)   /* is there a character in the buffer ? */
     return(0);
  tempch = (int) window->Table[window->tail];
  window->tail = (window->tail + 1) & (Console_limit - 1);
  if (debugflags & Keyboard_Flag)
   ServerDebug("Read char %x from console", tempch);
  return(tempch);
}

/**
*** Console_Read() is called when a Read request is received from the
*** transputer. It extracts the read parameters from the message and from the
*** current attributes and calls a suitable routine to deal with the request.
*** This leaves most of the work to be done by raw and cooked_console_read().
**/


void Console_Read(myco)
Conode *myco;
{ WORD count = (mcb->Control)[ReadSize_off];
  WORD timeout = (mcb->Control)[ReadTimeout_off];
  WORD timelimit;
  Window *window = (Window *) myco->extra;

  if (count eq 0L)          /* special case */
    { Request_Return(ReadRc_EOD, 0L, 0L);
      return;
    }
               /* I want a sensible limit on the amount read, 1 Kbyte reads */
               /* from a console are silly.                                 */
  if (count >= IOCDataMax) count = IOCDataMax - 1;

  if (timeout eq -1L)                        /* And work out the timelimit */
    timelimit = MAXTIME;
  else
    timelimit = Now + divlong( timeout, time_unit);

  if (IsAnAttribute(&(window->attr), ConsoleRawInput))
    raw_Console_read(myco, (int) count, timelimit);
  else
    cooked_Console_read(myco, (int) count, timelimit);
}

PRIVATE void send_console_data();
/**
*** In raw input mode, I read as many characters as I am asked for
*** or as there are available within the time limit. I perform no
*** editing of any sort, and echo all characters if the echo attribute is set.
**/

PRIVATE void raw_Console_read(myco, count, timelimit)
Conode *myco;
int    count;
WORD   timelimit;
{ BYTE *buff, local_buf[80];
  Window *window  = (Window *) myco->extra;
  WORD echo       = IsAnAttribute(&(window->attr), ConsoleEcho);
  WORD reply_port = mcb->MsgHdr.Reply;
  int  i, temp, ownbuf = 0;

                             /* Work out the buffer where to store the data */
  if (count < 80)
    buff = &(local_buf[0]);
  else
   { buff = malloc(count);
     if (buff eq (BYTE *) NULL)
      { Request_Return(EC_Error + SS_IOProc + EG_NoMemory + EO_Server, 0L, 0L);
        return;
      }
     ownbuf++;
   }
                          /* Usually I want to wait for a key, so I move to */
                          /* PollingCo straight away.                       */
  AddTail(Remove(&(myco->node)), PollingCo);
#if multi_tasking
  AddMultiwait(myco, Multi_ConsoleInput, window->handle);
#else
  myco->type = CoReady;
#endif

  for (i = 0; i < count; )
     { if (temp = window_getchar(window))
        { buff[i++] = (BYTE) temp;
          if (echo)       /* Should I echo the character ? */
            outputch(temp, window);
         }
       else
        { myco->timelimit = timelimit;
          Suspend();                       /* No, suspend coroutine for a bit. */

          if (myco->type eq CoSuicide)     /* Check for some special cases. */
           {  if (ownbuf) free(buff);
#if multi_tasking
              ClearMultiwait(myco, Multi_ConsoleInput, window->handle);
#endif
              (void) Console_TidyStream(myco);
              Seppuku();
           }
          elif (myco->type eq CoTimeout)          /* Time limit exceeded ? */
           {
             if (i > 0)                    /* Read any characters at all ? */
              { count = i; break; }
             else                           /* No, so return timeout. */
              { mcb->MsgHdr.Reply = reply_port;
                Request_Return(EC_Recover + SS_IOProc + EG_Timeout + EO_Stream,
                               0L, 0L);
                goto RawRead_end;
              }
           }
        }
     }

  send_console_data(reply_port, ReadRc_EOD, buff, (WORD) count);

RawRead_end:

  if (ownbuf) free(buff);
  PostInsert(Remove(&(myco->node)), Heliosnode); /* Finished, so wait for next request */
#if multi_tasking
  ClearMultiwait(myco, Multi_ConsoleInput, window->handle);
#endif
}

/**
***   Cooked input is rather more complicated. I am responsible
*** for all input processing, including cursor keys, backspace
*** and delete. Data cannot be sent until a carriage return has
*** been detected, because it might be erased. However, I must
*** echo it anyway so that the user can see what he is typing.
*** This means reading the data into a static buffer until
*** it is ready to be sent. Interleaving cooked reads is safe,
*** but I hate to think what will happen if cooked and raw
*** reads are mixed together.
***
*** As the data is read in and processed I store it in cooked_buffer, and
*** I use cooked_count to keep track of my current position in this buffer.
*** Routine fill_buffer() is responsible for actually getting and processing
*** the keys, returning one of Cooked_EOF, _EOF or _Done. There is a hideous
*** problem in that a particular cooked read may have extracted and processed
*** lots of data without satisfying the request, leaving the data in the
*** buffer. If flag outstanding is set then this is the case. The data still
*** in the buffer may satisfy the current Read request, leaving some data still
*** in the buffer or completely emptying the buffer.
***
*** If the Read has not yet been satisfied by now, I start to extract data from
*** the buffer and process it. If no data is available the coroutine must
*** suspend itself, partly to release the buffer and partly in order to let
*** poll_the_devices() be called because that is the routine which will put
*** more data into the buffer. As for the rest of the code, it is too hideous
*** to describe.
***
*** To handle the EOF condition properly, I must be able to send an EOF to a
*** stream even though it has already received a ReadRc_EOF. Hence I store
*** the last stream to EOF.
***
*** If a ctrl-C has been typed while processing the cooked read, the read must
*** abort with EOF. It is assumed that there will be only one cooked read
*** happening in a window at any one time, so the coroutine can store a pointer
*** in the window structure. When a ctrl-C is detected the event handler zaps
*** a flag in the conode structure, which is detected next time I poll.
**/

PRIVATE int fn( fill_buffer, (WORD, Window *));

PRIVATE void cooked_Console_read(myco, count, timelimit)
Conode *myco;
int    count;
WORD   timelimit;
{ BYTE *buff, local_buf[80];
  Window    *window = (Window *) myco->extra;
  Microwave *cooker = &(window->cooker);
  WORD echo         = IsAnAttribute(&(window->attr), ConsoleEcho);
  int  temp, ownbuf =0;
  WORD amount, port = mcb->MsgHdr.Reply;

  if (myco->id eq cooker->EOFco)  /* is the stream trying to read past ctrl-D */
   { Request_Return(ReadRc_EOF, 0L, 0L);
     return;
   }

                             /* Work out the buffer where to store the data */
  if (count < 80)
    buff = &(local_buf[0]);
  else
   { buff = malloc(count);
     if (buff eq (BYTE *) NULL)
      { Request_Return(EC_Error + SS_IOProc + EG_NoMemory + EO_Server, 0L, 0L);
        return;
      }
     ownbuf++;
   }
 
  if (cooker->outstanding)      /* did the previous read leave some data buffered ? */
   { if (count >= strlen(&(cooker->buffer[0])))     /* yes, so send it. */
      { strcpy(buff, &(cooker->buffer[0]));         /* leaving nothing in buffer */
                                   /* do not bother to wait for more data */
        send_console_data(port, ReadRc_EOD, buff, (WORD) strlen(buff));
        cooker->outstanding = 0; cooker->buffer[0] = '\0'; cooker->count = 0;
        if (ownbuf) free(buff);
        return;
      }

                                              /* yes, and it is > requested */
     send_console_data(port, ReadRc_EOD, &(cooker->buffer[0]),
                       (WORD) count);
     strcpy(&(cooker->buffer[0]), &(cooker->buffer[0]) + count);
                              /* lose the bit sent */
     if (ownbuf) free(buff);
     return;
   }
                          /* Usually I want to wait for a key, so I move to */
                          /* PollingCo straight away.                       */
  AddTail(Remove(&(myco->node)), PollingCo);
#if multi_tasking
  AddMultiwait(myco, Multi_ConsoleInput, window->handle);
#else
  myco->type = CoReady;
#endif

  window->cooker.readerco = myco;        /* I am reading from this window! */

  forever
    { temp = fill_buffer(echo, window);     /* fill_buffer does the editing */

junkgoto:

      if (temp eq Cooked_EOF)          /* Has ctrl-D been detected ? */
       { strcpy(buff, &(cooker->buffer[0]));  /* Yes,send whatever data is buffered */
         amount = strlen(buff);
         if (amount <= (WORD) count)
          { 
            send_console_data(port, ReadRc_EOF, buff, amount);
            cooker->count = 0; cooker->buffer[0] = '\0';      /* buffer empty */
            cooker->EOFco = myco->id;     /* stop the stream reading past EOF */
            break;
          }
         else
          { 
            send_console_data(port, ReadRc_EOF, buff, (WORD) count);
            strcpy(&(cooker->buffer[0]), &(cooker->buffer[0]) + count);
            cooker->outstanding = 1;
            break;
          }
       }

      if (temp eq Cooked_Done)             /* was there a carriage return ? */
        { strcpy(buff, &(cooker->buffer[0]));
          amount = strlen(buff);
          if (amount <= count)
           { 
             send_console_data(port, ReadRc_EOD, buff, amount);
             cooker->count = 0; cooker->buffer[0] = '\0';     /* buffer empty */
             break;
           }
          else
           { 
             send_console_data(port, ReadRc_EOD, buff, (WORD) count);
             strcpy(&(cooker->buffer[0]), &(cooker->buffer[0]) + count);
             cooker->outstanding = 1;
             break;
           }
        }

      myco->timelimit = timelimit;
      Suspend();                                /* wait a bit for more data */

      if (myco->type eq CoSuicide)
        { if (ownbuf) free(buff);
#if multi_tasking
          ClearMultiwait(myco, Multi_ConsoleInput, window->handle);
#endif
          window->cooker.readerco = (Conode *) NULL;
          (void) Console_TidyStream(myco);
          Seppuku();
        }

      if (myco->flags & CoFlag_CtrlCed)
       { temp = Cooked_EOF; goto junkgoto; }   /* Yukk !! */

      if (myco->type eq CoTimeout)        /* Has time limit been exceeded ? */
       { mcb->MsgHdr.Reply = port;
         Request_Return(EC_Recover + SS_IOProc + EG_Timeout + EO_Stream,
                        0L, 0L);
         break;
       }
                         /* Data is not yet ready, but there is still time. */
    }

  if (ownbuf) free(buff);
  window->cooker.readerco = (Conode *) NULL;
  PostInsert(Remove(&(myco->node)), Heliosnode); /* Finished, so wait for next request */
#if multi_tasking
  ClearMultiwait(myco, Multi_ConsoleInput, window->handle);
#endif
}



           /* Routine fill_buffer takes data from the main console circular */
           /* buffer and performs local editing.                            */
PRIVATE int fill_buffer(echo, window)
WORD echo;
Window *window;
{ int curr_ch;
  Microwave *cooker = &(window->cooker);
  char *buffer = &(cooker->buffer[0]);
  int  count   = cooker->count;

  while (curr_ch = window_getchar(window))
    switch(curr_ch)
      { case 0x0A :    /* treat carriage return and newline the same way... */
        case 0x0D : if (count eq Console_limit)
                     { if (echo)
                        window_output("\b\b  \b\b", window);
                       count -= 2;
                     }
                    buffer[count++] = '\n';
                    buffer[count++] = '\0';
                    if (echo) window_output("\r\n", window);
                    cooker->count = count;
                    return(Cooked_Done);

        case 0x04 : if (count eq Console_limit)
                     { if (echo)
                        window_output("\b \b", window);
                       count--;
                     }
                    buffer[count++] = '\0';     /*ctrl-D, EOF */
                    cooker->count = count;
                    return(Cooked_EOF);

        case 0x08 : if (count)              /* backspace if possible */
                     { count--;
                       if (echo)
                         window_output("\b \b", window);    /* do the erase */
                     }
                   break;

        case 0x7F : count = 0;             /* erase line */
                    if (echo)
                     { PRIVATE UBYTE clearline[4] =
                       { '\r', 0x9B, 0x4B, '\0' };
                       window_output((BYTE *)clearline, window);
                     }
                    break;

                   /* worry about the cursor keys later */

        default : if (isprint(curr_ch))
                   { if (count eq Console_limit) 
                      { if (echo)
                         window_output("\b \b", window);
                        count--;
                      }
                     buffer[count++] = (char) curr_ch;
                     if (echo) outputch(curr_ch, window);
                   }
      }

  cooker->count = count;
  return(Cooked_EOD);        /* no more characters in the buffer, so return */
}

/**
*** Write data to the console.
***
***  This is ridiculously complicated. The first problem is that there
*** are two types of write requests, single writes and multiple writes.
*** Console_Write takes care of this bit of the protocol, sending its
*** data to routine write_out. The routine also takes care of XON/XOFF,
*** and hence it is responsible for all the timing.
***
***  Data arrives at write_out() in fairly small lumps, so that the
*** server is not locked up when handling large writes to the console :
*** Console_write() suspends itself between lumps. There would be another
*** problem here if I allowed multiple messages, but I ensure that I only
*** ever get one data message. I just hope that nobody tries to write
*** more than 64K to the screen at any one time.
***
***  Writeout() checks which output mode the console is in at the moment
*** and performs raw/cooked editing. The data is then output via routine
*** output(). Output() is used elsewhere to write data to the screen.
*** There is another routine outputch() to output a single character, and
*** this is used mainly for echoing. In output(), on the ST the first
*** stage is to filter the data through ANSI_to_VT52() in basicio.c, to
*** convert any ANSI escape sequences to VT52 ones. After all this I am
*** finally ready to send my data to the screen.
***
***  Incidentally, the system would get extremely confused if anybody
***tried to output '\0', or if writes were interleaved in the wrong way.
**/

#if AMIGA
#define CONSOLE_WRITE_SIZE 255L
#else
#define CONSOLE_WRITE_SIZE 40L
#endif

PRIVATE void write_out();

void Console_Write(myco)          /* handle a write to screen request */
Conode *myco;
{ Window *window = (Window *) myco->extra;
  WORD rawmode = IsAnAttribute(&(window->attr), ConsoleRawOutput);
  WORD timeout = mcb->Control[WriteTimeout_off];
  WORD timelimit;
  WORD count   = mcb->Control[WriteSize_off];  /* amount of data to be written */
  WORD replyport = mcb->MsgHdr.Reply;
  int  newbuf=0;
  BYTE *buff, *a_ptr, local_buf[80];
  word amount, written, temp;

  if (count eq 0L)                    /* special case */
   { mcb->Control[Reply1_off] = 0L;
     Request_Return(WriteRc_Done, 1L, 0L);
     return;
   }

  if (count > 0x7FFFL)
   { Request_Return(EC_Error + SS_IOProc + EG_WrongSize + EO_File, 0L, 0L);
     return;
   }

  if (timeout eq -1L)
     timelimit = MAXTIME;
  else
     timelimit = Now + divlong( timeout, time_unit);

  if (count < 80)
   buff = &(local_buf[0]);
  else
   { buff = (BYTE *) malloc((uint) count);
     if (buff eq (BYTE *) NULL)
      { Request_Return(EC_Error + SS_IOProc + EG_NoMemory + EO_Server, 0L, 0L);
        return;
      }
     newbuf++;
   }

  if (mcb->MsgHdr.DataSize ne 0L)         /* has the data arrived already ? */
    { amount = (word) mcb->MsgHdr.DataSize;
      memcpy(buff, mcb->Data, mcb->MsgHdr.DataSize);
    }
  else                                     /* No, so I have to get the data */
    { WORD fetched   = 0L;
      
            /* The data is read in in lumps of upto MaxData, */
            /* and put into the same buffer. Once the buffer */
            /* has been filled I start printing it.          */

      mcb->Control[Reply1_off] = (count > maxdata) ? maxdata : count;
      mcb->Control[Reply2_off] = maxdata;
      mcb->MsgHdr.Flags = MsgHdr_Flags_preserve; /* must preserve the route here */
      Request_Return(WriteRc_Sizes, 2L, 0L);

      a_ptr = buff;

      while ((fetched < count) && (timelimit > Now))
       { myco->timelimit = timelimit;
         Suspend();                         /* Wait for the data to be sent */
         if ((myco->type eq CoSuicide) || (myco->type eq CoTimeout))
          { if (newbuf) free(buff);
            (void) Console_TidyStream(myco);
            Seppuku();
          }
         fetched   += (word) mcb->MsgHdr.DataSize;
         memcpy(a_ptr, mcb->Data, mcb->MsgHdr.DataSize);
         a_ptr     += mcb->MsgHdr.DataSize;           /* next bit of buffer */
         timelimit = Now + divlong(timeout, time_unit);     /* reset timer */
       }

                 /* Now either I have got all the data or I have timed out  */
                 /* Either way I want to print some of it, to prevent       */
                 /* infinite loops of Writing 0 bytes within timeout.       */
                 /* I want to display FETCHED chars */
      amount = fetched;
    }

        /* When I get here, all the data has arrived from the */
        /* transputer. I need to display AMOUNT characters    */
        /* starting at location buff, within TIMELIMIT.       */

        /* put myself into PollingCo, to avoid further messages */
   AddTail(Remove(&(myco->node)), PollingCo);
#if multi_tasking
   AddMultiwait(myco, Multi_ConsoleInput, window->handle);
#else
   myco->type = CoReady;
#endif
        /* I have a large outer loop, and each time around my loop */
        /* I check for XOFF and display CONSOLE_WRITE_SIZE chars.  */
        /* I use pointer currptr to keep track of how far I am in  */
        /* the buffer, and word written contains how much written  */

   a_ptr = buff; written = 0L;

   while ((written < amount) && (timelimit > Now))
    {                                       /* Start by worrying about XOFF */
     while (window->XOFF)
      { myco->timelimit = timelimit;
        Suspend();                              /* suspend myself for a bit */
        if (myco->type eq CoSuicide)
         { if (newbuf) free(buff);
#if multi_tasking
           ClearMultiwait(myco, Multi_ConsoleInput, window->handle);
#endif
           (void) Console_TidyStream(myco);
           Seppuku();
         }
        if (myco->type eq CoTimeout)
         { mcb->MsgHdr.Reply = replyport;
           Request_Return(EC_Recover + SS_IOProc + EG_Timeout + EO_Stream,
                          0L, 0L);
           goto end;
         }
      }

     temp = ((amount - written) < CONSOLE_WRITE_SIZE) ?
             (amount - written) : CONSOLE_WRITE_SIZE;
     write_out(a_ptr, temp, rawmode, window);
     a_ptr = a_ptr + temp;
     written += temp;

     myco->type    = CoReady;
     myco->timelimit = MAXINT;        
     Suspend();                                /* Release the CPU for a bit */
     if (myco->type eq CoSuicide)
      { if (newbuf) free(buff);
#if multi_tasking
        ClearMultiwait(myco, Multi_ConsoleInput, window->handle);
#endif
        (void) Console_TidyStream(myco);
        Seppuku();
      }      
   }

                /* When I get here, I have written WRITTEN */
                /* chars within the time available */
                /* So inform the transputer. */
  mcb->Control[Reply1_off] = written;
  mcb->MsgHdr.Reply = replyport;

  Request_Return(WriteRc_Done, 1L, 0L);
 
end:            /* tidy up everything */
  PostInsert(Remove(&(myco->node)), Heliosnode); /* ready for next message */
  if (newbuf) free(buff);
#if multi_tasking
  ClearMultiwait(myco, Multi_ConsoleInput, window->handle);
#endif
               /* Finished !!! */
}     


    /* Routine write_out() takes a fairly small amount of data, processes */
    /* it as per the current output mode (raw/cooked), and outputs it.    */
PRIVATE void write_out(data, amount, rawmode, window)
UBYTE *data;
WORD amount, rawmode;
Window *window;
{ BYTE *newdata = (BYTE *) &(misc_buffer1[0]);
  if (rawmode)
    while (amount--) *newdata++ = *data++;
  else
    { 
      while (amount--)
       { if (((32 <= *data) && (*data <= 255)) ||    /* printable chars */
             (( 0x07 <= *data) && (*data <= 0x0D)))  /* usual control chars */
           *newdata++ = *data;
         if (*data eq '\n') *newdata++ = '\r';
         data++;
       }
    }

  *newdata = '\0';
  window_output(&(misc_buffer1[0]), window);
}



/**
*** This is a little routine to send back a reply to a console read request.
*** The console routines usually involve polling, so the message buffer will
*** be corrupted.
**/

PRIVATE void send_console_data(port, fnrc, buffer, amount)
WORD port, fnrc, amount;
BYTE *buffer;
{ BYTE *tempptr = mcb->Data;
  mcb->MsgHdr.Reply = port;
  mcb->MsgHdr.Dest  = NullPort;     /* Seems to be needed, not sure why */
  mcb->Data = buffer;

  Request_Return(fnrc, 0L, amount); 
  mcb->Data = tempptr;
}

            /* Output a single character, used mainly for echoing */
            /* I want to go via output(), to handle ANSI-VT52     */
            /* conversions etc, even though it seems a little bit */
            /* silly.                                             */
void outputch(ch, window)
int ch;
Window *window;
{ BYTE temp[4];
  temp[0] = (BYTE) ch; temp[1] = '\0';
  window_output(&(temp[0]), window);
}

char *lhex_out(dest, val, width)
char *dest;
UWORD val;
int width;
{ int i, j, k;

  if ((val eq 0L) && (width eq -1)) { *dest++ = '0'; return(dest); }

  if (width eq -1)
   { for (i = 7; i > 0; i--)   /* Work out most significant nibble */
     if (((val >> (i << 2)) & 0x0F) ne 0)
      break;
   }
  else
   i = width - 1;

  for (j = 0; j <= i; j++)
   { k = (int) ((val >> ((i - j) << 2)) & 0x0F);
     if (k > 9)
      *dest++ = (char) ('a' + k - 0x0a);
     else
      *dest++ = (char) ('0' + k);
   }

  return(dest);     
}

char *lint_out(dest, val)
char *dest;
WORD val;
{ static UWORD tab[9] = { 1000000000L, 100000000L, 1000000L, 100000L,
                         10000L, 1000L, 100L, 10L, 1L };
  int i, k;

  if (val eq 0L) { *dest++ = '0'; return(dest); } 
  if (val < 0L)
   { *dest++ = '-'; val = -1L * val; }

  for (i = 0; val < tab[i]; i++);

  for ( ; i < 9; i++)
   { k = 0;
     while (val >= tab[i]) { val -= tab[i]; k++; }
     *dest++ = (char) ('0' + k);
   }

  return(dest);
}

#if ANSI_prototypes

void ServerDebug(text, ...)
char *text;
{ char *dest = &(err_buff[0]);
  int  longs, width;
  va_list args;

  va_start(args, text);

  for ( ; *text ne '\0'; text++)
   if (*text ne '%')
    *dest++ = *text;
   else
    { longs = 0; width = -1;
      again:
      switch (*++text)
       { case '%' : *dest++ = '%'; break;
         case 'l' :
         case 'L' : longs = 1; goto again;
         case '0' : case '1' : case '2' : case '3' : case '4' :
         case '5' : case '6' : case '7' : case '8' : case '9' :
                    width = *text - '0';
                    goto again;
         case 'c' :
         case 'C' : *dest++ = (char) va_arg(args, int); break;
         case 'p' :
         case 'P' : longs = 1;
         case 'x' :
         case 'X' : if (longs)
                       dest = lhex_out(dest, va_arg(args, UWORD), width);
                    else
                     dest = lhex_out(dest,(UWORD) va_arg(args, uint), width);
                    break;
         case 'd' :
         case 'D' : if (longs)
                       dest = lint_out(dest, va_arg(args, WORD));
                    else
                     dest = lint_out(dest, (WORD) va_arg(args, int)); 
                    break;
         case 's' :
         case 'S' : { char *dat = va_arg(args, char *);
                      int i = 0;
                      for ( ; (*dat ne '\0') && (i < 80);i++ )
                       *dest++ = *dat++;
                      break;
                    }
         case 'q' :
         case 'Q' : goto done;
       }
    }

  *dest++ = '\r'; *dest++ = '\n';

done: 
  *dest = '\0';

  if ((log_dest eq Log_to_screen) || (log_dest eq Log_to_both))
    output(err_buff);
  if ((log_dest eq Log_to_file) || (log_dest eq Log_to_both))
    write_to_log(err_buff);

  va_end(args);
}

#else

void ServerDebug(text, x)
char *text;
int x;
{ char *dest = &(err_buff[0]);
  int  longs, width;
  int  *args_ptr = &x;

  for ( ; *text ne '\0'; text++)
   if (*text ne '%')
    *dest++ = *text;
   else
    { longs = 0; width = -1;
      again:
      switch (*++text)
       { case '%' : *dest++ = '%'; break;
         case 'l' :
         case 'L' : longs = 1; goto again;
         case '0' : case '1' : case '2' : case '3' : case '4' :
         case '5' : case '6' : case '7' : case '8' : case '9' :
                    width = *text - '0';
                    goto again;
         case 'c' :
         case 'C' : *dest++ = (char) (*args_ptr++); break;
         case 'p' :
         case 'P' : longs = 1;
         case 'x' :
         case 'X' : if (longs)
                     { UWORD *tmpptr = (UWORD *) args_ptr;
                       dest = lhex_out(dest, *tmpptr++, width);
                       args_ptr = (int *) tmpptr;
                     }
                    else
                     dest = lhex_out(dest,(UWORD) (*args_ptr++), width);
                    break;
         case 'd' :
         case 'D' : if (longs)
                     { WORD *tmpptr = (WORD *) args_ptr;
                       dest = lint_out(dest, *tmpptr++);
                       args_ptr = (int *) tmpptr;
                     }
                    else
                     dest = lint_out(dest, (WORD) (*args_ptr++));
                    break;
         case 's' :
         case 'S' : { char **str_ptr = (char **) args_ptr;
                      char *dat;
                      dat = *str_ptr++;
                      args_ptr = (int *) str_ptr;
                      for ( ; *dat ne '\0'; )
                       *dest++ = *dat++;
                      break;
                    }
         case 'q' :
         case 'Q' : goto done;
       }
    }

  *dest++ = '\r'; *dest++ = '\n';

done: 
  *dest = '\0';

  if ((log_dest eq Log_to_screen) || (log_dest eq Log_to_both))
    output(err_buff);
  if ((log_dest eq Log_to_file) || (log_dest eq Log_to_both))
    write_to_log(err_buff);
}

#endif  /* ANSI_prototypes */

/**---------------------------------------------------------------------------
***
*** That was the general code for the window and console devices. The code
*** below is an ANSI screen emulator. Most of the code is used to analyse
*** the input data, and maintain a shadow screen. Then there are a few
*** routines which actually have to do the work on the real screen.
*** I choose to have separate sets of routines for the different devices,
*** so these sets should probably be moved to the local modules at some
*** stage.
***
***---------------------------------------------------------------------------
**/

#if use_ANSI_emulator

/**
*** A general purpose ANSI terminal emulator, which can be easily modified
*** to deal with particular terminals.
**/
#define Mode_Plain      0
#define Mode_Bold       0x0001
#define Mode_Italic     0x0002
#define Mode_Underline  0x0004
#define Mode_Inverse    0x0008

/**
*** Here are the routines corresponding to the ANSI sequences.
**/
PRIVATE void fn( add_ch,            (int));
PRIVATE void fn( insert_chars,      (int));
PRIVATE void fn( cursor_up,         (int));
PRIVATE void fn( cursor_down,       (int));
PRIVATE void fn( cursor_right,      (int));
PRIVATE void fn( cursor_left,       (int));
PRIVATE void fn( cursor_next,       (int));
PRIVATE void fn( cursor_prev,       (int));
PRIVATE void fn( move_cursor,       (int, int));
PRIVATE void fn( erase_endofscreen, (void));
PRIVATE void fn( erase_endofline,   (void));
PRIVATE void fn( insert_line,       (void));
PRIVATE void fn( delete_line,       (void));
PRIVATE void fn( delete_char,       (int));
PRIVATE void fn( scroll_up,         (int));
PRIVATE void fn( scroll_down,       (int));
PRIVATE void fn( set_rendition,     (Screen *));

/**
*** These are the functions which have to do the work. They are described below.
**/
PRIVATE void fn( clear_screen,      (void));
PRIVATE void fn( clear_eol,         (void));
PRIVATE void fn( move_to,           (int, int));
PRIVATE void fn( refresh_screen,    (int, int));
PRIVATE void fn( backspace,         (void));
PRIVATE void fn( carriage_return,   (void));
PRIVATE void fn( linefeed,          (void));
PRIVATE void fn( set_mark,          (void));
PRIVATE void fn( use_mark,          (void));
PRIVATE void fn( foreground,        (int));
PRIVATE void fn( background,        (int));
PRIVATE void fn( set_bold,          (int));
PRIVATE void fn( set_italic,        (int));
PRIVATE void fn( set_underline,     (int));
PRIVATE void fn( set_inverse,       (int));
PRIVATE void fn( ring_bell,         (void));

/**
*** These routines are responsible for actually sending characters to the output
*** device. They may be defined as macros.
**/
#ifndef send_ch
PRIVATE void fn( send_ch, ());
#endif
#ifndef flush
PRIVATE void fn( flush, ());
#endif

/**
*** Init_Ansi() should be called during the initialisation process. Note that
*** it clears the screen, so that I know exactly what is on the screen when I
*** start up.
**/

PRIVATE void fn( reset_escape, (Screen *));

WORD Init_Ansi(window, No_Rows, No_Cols)
Window *window;
int No_Rows, No_Cols;
{ int i, j;
  BYTE **map;
  Screen *screen = &(window->screen);

  if ((map = (BYTE **) malloc(No_Rows * sizeof(BYTE *))) eq (BYTE **) NULL)
    return(FALSE);
  for (i = 0; i < No_Rows; i++) 
   if ((map[i] = (BYTE *) malloc(No_Cols)) eq (BYTE *) NULL)
    { for (j = 0; j < i; j++) free(map[j]);
      free(map);
      return(FALSE);
    }

  screen->map    = map;
  screen->Rows   = No_Rows;
  screen->Cols   = No_Cols;
  screen->mode   = Mode_Plain;
  reset_escape(screen);

  { static char cls[2] = { 0x0C, 0x00 };
    ANSI_out(&(cls[0]), window);
  }
  return(TRUE);
}

void Tidy_Ansi(screen)
Screen *screen;
{ int i;
  BYTE **map = screen->map;

  for (i = 0; i < screen->Rows; i++) free(map[i]);
  free(map);
}

/**
*** This resets a window's escape processing variables.
**/
PRIVATE void reset_escape(screen)
Screen *screen;
{ screen->escape_found = 0; screen->in_escape = 0;
  screen->args[0] = 0; screen->args[1] = 0; screen->args[2] = 0;
  screen->args[3] = 0; screen->args[4] = 0;
  screen->args_ptr = &(screen->args[0]); screen->gotdigits = 0;
  screen->firstdigit = 1;
}

/**
*** When I get text for a particular window, I always put the relevant
*** variables into statics rather than access them via the window pointer.
*** This should speed things up a bit.
**/

PRIVATE int    Cur_x, Cur_y, Rows, Cols, Screen_mode;
PRIVATE BYTE   **map;
PRIVATE Window *current_window;
PRIVATE WORD   handle;

void ANSI_out(str, window)
STRING str;
Window *window;
{ Screen *screen = &(window->screen);
  Cur_x          = screen->Cur_x;
  Cur_y          = screen->Cur_y;
  map            = screen->map;
  Rows           = screen->Rows;
  Cols           = screen->Cols;
  Screen_mode    = screen->mode;
  current_window = window;
  handle         = window->handle;

  for ( ; *str ne '\0'; str++)
   { if (screen->in_escape)
      switch(*str)
       {        /* If it is a digit, it is part of the current arg */
         case '0': case '1': case '2': case '3': case '4':
         case '5': case '6': case '7': case '8': case '9':
           *(screen->args_ptr) = (10 * *(screen->args_ptr)) + (*str - '0');
           if (screen->firstdigit)
             { screen->gotdigits++; screen->firstdigit = 0; }
           break;

                /* A semicolon marks the end of the current arg. */
     case ';':  screen->args_ptr++; screen->firstdigit = 1; break;

     case '@':  insert_chars((screen->gotdigits) ? screen->args[0] : 1);
                goto end;

     case 'A':  cursor_up((screen->gotdigits) ? screen->args[0] : 1);
                goto end;

     case 'B':  cursor_down((screen->gotdigits) ? screen->args[0] : 1);
                goto end;

     case 'C':  cursor_right((screen->gotdigits) ? screen->args[0] : 1);
                goto end;

     case 'D':  cursor_left((screen->gotdigits) ? screen->args[0] : 1);
                goto end;

     case 'E':  cursor_next((screen->gotdigits) ? screen->args[0] : 1);
                goto end;

     case 'F':  cursor_prev((screen->gotdigits) ? screen->args[0] : 1);
                goto end;

     case 'H':  move_cursor( (screen->gotdigits > 0) ? (screen->args[0] - 1) : 0,
                         (screen->gotdigits > 1) ? (screen->args[1] - 1) : 0);
                goto end;

     case 'J':  erase_endofscreen(); goto end;

     case 'K':  erase_endofline(); goto end;

     case 'L':  insert_line(); goto end;

     case 'M':  delete_line(); goto end;

     case 'P':  delete_char((screen->gotdigits) ? screen->args[0] : 1);
                goto end;

     case 'S':  scroll_up((screen->gotdigits) ? screen->args[0] : 1);
                goto end;

     case 'T':  scroll_down((screen->gotdigits) ? screen->args[0] : 1);
                goto end;

     case 'm':  set_rendition(screen);
     default :

end:
                reset_escape(screen);
       }

     else   /* Not in escape sequence... yet */
      { if (screen->escape_found && (*str eq '['))
         { screen->in_escape = TRUE; screen->escape_found = FALSE; continue; }

        screen->escape_found = FALSE;
        switch(*str)
         {      /* Line feed only : record terminator should have */
                /* expanded already. */
           case '\n' :  linefeed(); break;

                /* Single carriage return */
           case '\r' :  carriage_return(); break;

                /* Bell */
           case 0x07 :  ring_bell(); break;

                /* Form Feed */
           case 0x0C :  clear_screen(); break;

                /* Backspace */
           case 0x08 :  backspace(); break;

               /* Vertical tab */
           case 0x0B :  cursor_up(1); break;

               /* CSI */
           case 0x009B :
           case 0xFF9B : screen->in_escape = TRUE; break;

               /* possibly part of CSI */
           case 0x1B : screen->escape_found = TRUE; break;

              /* Horizontal tab */
           case 0x09 : { int i = 8 - (Cur_x % 8);
                         if ((Cur_x + i) >= Cols)
                          { carriage_return(); linefeed(); }
                         else
                          move_to(Cur_y, Cur_x + i);
                         break;
                       }

              /* Something else - is it printable ? */
           default    : if (*str >= ' ') add_ch(*str);
        }
      }
   }

  flush();
  screen->Cur_x = Cur_x;
  screen->Cur_y = Cur_y;
  screen->mode  = Screen_mode;
}

/**
***
**/

PRIVATE void insert_chars(count)
int count;
{ int i;

  if ((Cur_x + count) >= Cols)
   count = Cols - Cur_x;

  set_mark();

  for (i = 0; i < count; i++)
    send_ch(' ');
  for (i = Cur_x; i < (Cols - count); i++)
    send_ch(map[Cur_y][i]);
  for (i = Cols - 1; i >= (Cur_x + count); i--)
    map[Cur_y][i] = map[Cur_y][i - count];
  for (i = 0; i < count; i++)
    map[Cur_y][Cur_x+i] = ' ';
    
  use_mark();
}

PRIVATE void cursor_up(count)
int count;
{ Cur_y -= count;
  if (Cur_y < 0) Cur_y = 0;
  move_to(Cur_y, Cur_x);
}

PRIVATE void cursor_down(count)
int count;
{ 
  Cur_y += count;
  if (Cur_y >= Rows) Cur_y = Rows - 1;

  move_to(Cur_y, Cur_x);
}

PRIVATE void cursor_right(count)
int count;
{
  Cur_x += count;
  if (Cur_x >= Cols) Cur_x = Cols - 1;

  move_to(Cur_y, Cur_x);
}

PRIVATE void cursor_left(count)
int count;
{ Cur_x -= count;
  if (Cur_x < 0) Cur_x = 0;

  if (Cur_x eq 0)
   carriage_return();
  else
   move_to(Cur_y, Cur_x);
}

PRIVATE void cursor_next(count)
int count;
{ int i;
  carriage_return();
  for (i = 0; i < count; i++) linefeed();
}

PRIVATE void cursor_prev(count)
int count;
{ Cur_y -= count; Cur_x = 0;
  if (Cur_y < 0) Cur_y = 0;
  move_to(Cur_y, Cur_x);
}

PRIVATE void move_cursor(y, x)
int y,x;
{ if (x < 0) x = 0;
  if (x >= Cols) x = Cols - 1;
  if (y < 0) y = 0;
  if (y >= Rows) y = Rows - 1;
  move_to(y, x);
}

PRIVATE void erase_endofscreen()
{ int i, j;

  set_mark();

  for (i = Cur_y; i < Rows; i++)
   for (j = (i eq Cur_y) ? Cur_x : 0 ; j < Cols; j++)
    map[i][j] = ' ';

  if ((Cur_x eq 0) && (Cur_y eq 0))
    clear_screen();
  else
    refresh_screen(Cur_y, Rows - 1);

  use_mark();
}

PRIVATE void erase_endofline()
{ int i;

  for (i = Cur_x; i < Cols; i++)
   map[Cur_y][i] = ' ';

  clear_eol();
}

PRIVATE void insert_line()
{ int i; BYTE *ptr = map[Rows-1];

  set_mark();

  for (i = Rows - 1; i > Cur_y; i--)
   map[i] = map[i-1];
  map[Cur_y] = ptr;
  for (i = 0; i < Cols; i++) ptr[i] = ' ';  
    
  refresh_screen(Cur_y, Rows - 1);  

  use_mark();
}

PRIVATE void delete_line()
{ BYTE *ptr = map[Cur_y];
  int i;

  set_mark();
  
  for (i = Cur_y; i < (Rows - 1); i++)
   map[i] = map[i+1];

  map[Rows - 1] = ptr;

  for (i = 0; i < Cols; i++) ptr[i] = ' ';

  refresh_screen(Cur_y, Rows - 1);

  use_mark();
}

PRIVATE void delete_char(count)
int count;
{ int i; char ch;

  set_mark();

  if ((Cur_x + count) > Cols) count = Cols - Cur_x;
  
  for (i = Cur_x; i < (Cols - count); i++)
   { ch = map[Cur_y][i + count];
     send_ch(ch); map[Cur_y][i] = ch;
   }
  
  for (i = 0; i < count; i++)
   { send_ch(' '); map[Cur_y][Cols - (i + 1)] = ' '; }
   
  use_mark();
}

PRIVATE void scroll_up(count)
int count;
{ int i;

  set_mark();
  if (count >= Rows) { clear_screen(); use_mark(); return; }

  move_to(Rows - 1, 0);
  for (i = 0; i < count; i++) linefeed();
  use_mark();
}

PRIVATE void scroll_down(count)
int count;
{ int i, j;
  BYTE *tempptr;
  
  set_mark();
  
  if (count >= Rows)
   clear_screen();
  else
   { for (i = 0; i < count; i++)
      { tempptr = map[Rows - 1];
        for (j = Rows - 1; j > 0; j--)
          map[j] = map[j-1];
        map[0] = tempptr;
      }
      
     for (i = 0; i < count; i++)
       for (j = 0; j < Cols; j++)
         map[i][j] = ' ';
     refresh_screen(0, Rows - 1);
   }
   
  use_mark();
}

PRIVATE void set_rendition(screen)
Screen *screen;
{ int i;

  for (i = 0; i < screen->gotdigits; i++)
   { if ( (screen->args[i] >= 30) && (screen->args[i] <= 37))
      foreground(screen->args[i] - 30);
     elif ( (screen->args[i] >= 40) && (screen->args[i] <= 47))
      background(screen->args[i] - 30);
     elif (screen->args[i] eq 7)
      { if (!(Screen_mode & Mode_Inverse))
         { set_inverse(1); Screen_mode |= Mode_Inverse; }
      }
     elif (screen->args[i] eq 4)
      { if (!(Screen_mode & Mode_Underline))
         { set_underline(1); Screen_mode |= Mode_Underline; }
      }
     elif (screen->args[i] eq 3)
      { if (!(Screen_mode & Mode_Italic))
         { set_italic(1); Screen_mode |= Mode_Italic; }
      }
     elif (screen->args[i] eq 1)
      { if (!(Screen_mode & Mode_Bold))
         { set_bold(1); Screen_mode |= Mode_Bold; }
      }
     elif (screen->args[i] eq 0)
      { if (Screen_mode & Mode_Bold)
         { set_bold(0); Screen_mode &= ~Mode_Bold; }
        if (Screen_mode & Mode_Italic)
         { set_italic(0); Screen_mode &= ~Mode_Italic; }
        if (Screen_mode & Mode_Underline)
         { set_underline(0); Screen_mode &= ~Mode_Underline; }
        if (Screen_mode & Mode_Inverse)
         { set_inverse(0); Screen_mode &= ~Mode_Inverse; }
      }
   }
}

PRIVATE void add_ch(ch)
int ch;
{ if (Cur_x eq Cols)     /* Implicit wrap-around */
   { carriage_return(); linefeed(); }

  map[Cur_y][Cur_x] = (BYTE) ch; send_ch(ch); Cur_x++;
}

/**
*** Here are the terminal-dependant routines. I start with some routines for
*** the PC, which are slightly unusual because everything is done in terms of
*** the video bios calls. Then I supply a version for the ST. Some of the
*** routines have to do work on the shadow screen as well as on the real
*** screen. However, before doing anything to the real screen I call
*** check_window(), which determines whether or not the relevant window is
*** actually active.
**/

#if PC
PUBLIC void fn( vbios_cls,        (void));
PUBLIC void fn( vbios_movecursor, (int, int));
PUBLIC void fn( vbios_outputch,   (int));
PUBLIC void fn( vbios_scroll,     (void));
PUBLIC void fn( vbios_bell,       (void));
PUBLIC int  fn( check_window,     (WORD));

/**
*** This routine should clear the entire screen and move the cursor to the
*** top-left corner.
**/

PRIVATE void clear_screen()
{ int i, j;

  for (i = 0; i < Rows; i++)
   for (j = 0; j < Cols; j++)
    map[i][j] = ' ';
  Cur_x = 0; Cur_y = 0;

  if (check_window(handle))
      vbios_cls();
}

/**
*** This routine should clear the rest of the current line, including the
*** cursor position, leaving the cursor position unchanged.
**/
PRIVATE void clear_eol()
{
  if (check_window(handle))
    { int i;
      for (i = Cur_x; i < Cols; i++) vbios_outputch(' ');
      vbios_movecursor(Cur_y, Cur_x);
    }
}

/**
*** This routine should move the cursor to the position specified.
**/

PRIVATE void move_to(y, x)
int y, x;
{ Cur_x = x; Cur_y = y;

   if (check_window(handle))
     vbios_movecursor(Cur_y, Cur_x);
}

/**
*** This routine should refresh the screen from the top left, leaving
*** the cursor position unchanged.
**/

PRIVATE void refresh_screen(start_y, end_y)
{  int i, j;

   if (!check_window(handle)) return;

   vbios_movecursor(start_y, 0);
   
   for (i = start_y; i <= end_y; i++)
    { 
      for ( j = 0; j < Cols; j++)
        vbios_outputch(map[i][j]);

      if (i < (Rows - 1))
       vbios_movecursor(i+1, 0);
    }  
}

/**
*** This routine should cause an audio or visual alert of some sort.
**/
PRIVATE void ring_bell()
{
  if (check_window(handle))
    vbios_bell();  
}

/**
*** This routine should move the cursor left one position, without erasing the
*** character.
**/
PRIVATE void backspace()
{ if (Cur_x eq 0) return;
  Cur_x --;

  if (check_window(handle))
    vbios_movecursor(Cur_y, Cur_x);
}

/**
*** This routine should move the cursor to the first column of the current row.
**/

PRIVATE void carriage_return()
{ Cur_x = 0;

  if (check_window(handle))
    vbios_movecursor(Cur_y, Cur_x);
}

/**
*** This routine should move the cursor down one row. If the cursor is on the
*** bottom row the screen should scroll up one line, and the actual cursor
*** position will remain unchanged. The cursor should NOT move to the beginning
*** of the line.
**/

PRIVATE void linefeed()
{ if (Cur_y >= (Rows - 1))
   { int i; BYTE *ptr = map[0]; /* Scroll up */
     for (i = 0; i < (Rows - 1); i++)
      map[i] = map[i+1];
     for (i = 0; i < Cols; i++)
      ptr[i] = ' ';
     map[Rows - 1] = ptr;
     Cur_y = Rows - 1;

     if (check_window(handle))
      { vbios_scroll();
        vbios_movecursor(Cur_y, Cur_x);
      }
     return;
   }
  else
   { Cur_y++;

     if (check_window(handle))
      vbios_movecursor(Cur_y, Cur_x);
   }
}

/**
*** set_mark() and use_mark() are used to remember and restore the current
*** screen position. Many terminal emulators provide a built-in facility
*** for this.
**/

PRIVATE int mark_x, mark_y;

PRIVATE void set_mark()
{
   mark_x = Cur_x; mark_y = Cur_y;
}

PRIVATE void use_mark()
{
  move_to(mark_y, mark_x);
}

/**
*** All the different attributes.
**/
extern BYTE vbios_attr;

PRIVATE void foreground(colour)
int colour;
{ colour = colour;
}

PRIVATE void background(colour)
int colour;
{ colour = colour;
}

PRIVATE void set_inverse(flag)
int flag;
{ if (check_window(handle))
   { if (flag)
      vbios_attr = 0x70;
     else
      vbios_attr = 0x07;
   }
}

PRIVATE void set_bold(flag)
int flag;
{ flag = flag;
}

PRIVATE void set_underline(flag)
int flag;
{ flag = flag;
}

PRIVATE void set_italic(flag)
int flag;
{ flag = flag;
}

/**
*** These routines are responsible for sending converted terminal output to
*** the screen, by one means or another. If there are multiple windows, I
*** always buffer the data and send it in one lump using send_to_window().
*** Otherwise, on the PC and the ST send_ch is #define'd to be an
*** immediate operation so there is no need to worry about flush().
**/

PRIVATE void send_ch(ch)
int ch;
{
  if (check_window(handle))
    vbios_outputch(ch);
}

PRIVATE void flush()
{
}

/**
*** The following code is used to switch screens.
**/
void redraw_screen(window)
Window *window;
{ Screen *screen = &(window->screen);
  char *text = &(window->node.direntry.Name[0]);
  BYTE **map = screen->map;
  int i, j, k;

  handle = window->handle;

  vbios_cls();

  for (i = 0; i < Rows; i++)
   { for (j = Cols - 1; j >= 0; j--)
       if (map[i][j] ne ' ') break;
     for (k = 0; k <= j; k++)
      vbios_outputch(map[i][k]);

     if (i < (Rows - 1))
      vbios_movecursor(i + 1, 0);
   }

  vbios_movecursor(0, Cols - strlen(text));
  set_inverse(1);
  for ( ; *text ne '\0'; text++)
   vbios_outputch(*text);
  set_inverse(0);
  vbios_movecursor(screen->Cur_y, screen->Cur_x);
}

#endif /* PC specific routines */


#if ST
/**
*** This routine should clear the entire screen and move the cursor to the
*** top-left corner.
**/

PRIVATE void clear_screen()
{ int i, j;
  for (i = 0; i < Rows; i++)
   for (j = 0; j < Cols; j++)
    map[i][j] = ' ';
  Cur_x = 0; Cur_y = 0;

  if (check_window(handle))
   { Bconout(2, 0x1B); Bconout(2, 'E'); }
}

/**
*** This routine should clear the rest of the current line, including the
*** cursor position, leaving the cursor position unchanged.
**/
PRIVATE void clear_eol()
{ if (check_window(handle))
   { Bconout(2, 0x1B); Bconout(2, 'K'); }
}

/**
*** This routine should move the cursor to the position specified.
**/

PRIVATE void move_to(y, x)
int y, x;
{ Cur_y = y; Cur_x = x;

  if (check_window(handle))
   { Bconout(2, 0x1B); Bconout(2, 'Y'); Bconout(2, 32+y); Bconout(2, 32+x); }
}

/**
*** This routine should refresh the screen from the top left, leaving
*** the cursor position unchanged.
**/

PRIVATE void refresh_screen(start_y, end_y)
{  int i, j;

   if (!check_window(handle)) return;

   Bconout(2, 0x1B); Bconout(2, 'f');    /* Disable cursor */
   Bconout(2, 0x1B); Bconout(2, 'Y'); Bconout(2 ,start_y + 32);
   Bconout(2, 32);
   
   for (i = start_y; i <= end_y; i++)
    { 
      for ( j = 0; j < Cols; j++)
        Bconout(2, map[i][j]);

      if (i < (Rows - 1))
       { Bconout(2, 0x1B); Bconout(2, 'Y'); Bconout(2, 32 + i + 1);
         Bconout(2, 32);
       }
    }  

  Bconout(2, 0x1B); Bconout(2, 'e');       /* Enable cursor */
}

/**
*** This routine should cause an audio or visual alert of some sort.
**/
PRIVATE void ring_bell()
{ PRIVATE BYTE beep_sound[] =
{ 0x00, 0x34, 0x01, 0x00, 0x02, 0x00, 0x03, 0x00, 0x04, 0x00, 0x05, 0x00,
  0x06, 0x00, 0x07, 0xFE, 0x08, 0x10, 0x09, 0x00, 0x0A, 0x00, 0x0B, 0x00,
  0x0C, 0x10, 0x0D, 0x09, 0xFF, 0x00
};

  if (check_window(handle))
   Dosound(beep_sound);
}

/**
*** This routine should move the cursor left one position, without erasing the
*** character.
**/
PRIVATE void backspace()
{ if (Cur_x eq 0) return;
  Cur_x --;
  if (check_window(handle))
    Bconout(2, 0x08);
}

/**
*** This routine should move the cursor to the first column of the current row.
**/

PRIVATE void carriage_return()
{ Cur_x = 0;
  if (check_window(handle))
    Bconout(2, 0x0D);
}

/**
*** This routine should move the cursor down one row. If the cursor is on the
*** bottom row the screen should scroll up one line, and the actual cursor
*** position will remain unchanged. The cursor should NOT move to the beginning
*** of the line.
**/

PRIVATE void linefeed()
{ if (Cur_y >= (Rows - 1))
   { int i; BYTE *ptr = map[0]; /* Scroll up */
     for (i = 0; i < (Rows - 1); i++)
      map[i] = map[i+1];
     for (i = 0; i < Cols; i++)
      ptr[i] = ' ';
     map[Rows - 1] = ptr;
     Cur_y = Rows - 1;
   }
  else
   Cur_y++;

  if (check_window(handle))  
    Bconout(2, 0x0A);
}

/**
*** set_mark() and use_mark() are used to remember and restore the current
*** screen position. Many terminal emulators provide a built-in facility
*** for this.
**/

PRIVATE int mark_x, mark_y;

PRIVATE void set_mark()
{
   mark_x = Cur_x; mark_y = Cur_y;
}

PRIVATE void use_mark()
{
  move_to(mark_y, mark_x);
}

PRIVATE void foreground(colour)
int colour;
{ colour = colour;
}

PRIVATE void background(colour)
int colour;
{ colour = colour;
}

PRIVATE void set_inverse(flag)
int flag;
{ if (check_window(handle))
   { Bconout(2, 0x1B); Bconout(2, flag ? 'p' : 'q'); }
}

PRIVATE void set_bold(flag)
int flag;
{ flag = flag;
}

PRIVATE void set_underline(flag)
int flag;
{ flag = flag;
}

PRIVATE void set_italic(flag)
int flag;
{ flag = flag;
}

/**
*** These routines are responsible for sending converted terminal output to
*** the screen, by one means or another. If there are multiple windows, I
*** always buffer the data and send it in one lump using send_to_window().
*** Otherwise, on the PC and the ST send_ch is #define'd to be an
*** immediate operation so there is no need to worry about flush().
**/

#if multiple_windows

PRIVATE void send_ch(ch)
int ch;
{ if (check_window(handle))
   Bconout(2, ch);
}
#endif

PRIVATE void flush()
{
}

/**
*** The following code is used to switch screens.
**/
void redraw_screen(window)
Window *window;
{ Screen *screen = &(window->screen);
  char *text = &(window->node.direntry.Name[0]);
  BYTE **map = screen->map;
  int i, j, k;

  handle = window->handle;

  Bconout(2, 0x1B); Bconout(2, 'f');    /* Disable cursor */
  Bconout(2, 0x1B); Bconout(2, 'E');    /* Cls */ 

  for (i = 0; i < Rows; i++)
   { for (j = Cols - 1; j >= 0; j--)
       if (map[i][j] ne ' ') break;
     for (k = 0; k <= j; k++)
      Bconout(2, map[i][k]);

     Bconout(2, '\r');
     if (i < (Rows - 1))
      Bconout(2, '\n');
   }

  move_to(0, Cols - strlen(text));
  set_inverse(1);
  for ( ; *text ne '\0'; text++)
   Bconout(2, *text);
  set_inverse(0);
  move_to(screen->Cur_y, screen->Cur_x);
  Bconout(2, 0x1B); Bconout(2, 'e');    /* Enable cursor */
}

#endif /* ST specific routines */

#endif /* use_ANSI_emulator */

