/* polygon.c	: Routines for handling polygons
 *
 * Copyright	:	TransTiriel	(c)1990
 *
 * Author	:	M.J. Charnley-Fisher
 *
 * Revisions	:
 *
 *	1.0A	10Mar90	MJCF	Created
 *
 * Description	:
 *
 * Most documented polygon filling algorithms use an active edge list by scan line 
 * in order to sort out where to draw each point. They then draw each pixel base on
 * the calculated scan lines. Whilst suitable for the T800, this is not ideal for
 * Blossom since blossom can blit trapezoid shaped areas. These routines are therefore
 * a modification which does it on an area basis - it is therefore considerably 
 * quicker (by a factor of between 1.3 and 15, depending on the average size of the
 * polygon)
 *
 * -------------------------------------------------------------------------------- *
 */

#include <stdio.h>
#include <stdlib.h>
#include <codes.h>

#include <glocal.h>
#include <graphics.h>

/* Local Structure used to store the active edges */
typedef struct Edge {
	struct Edge	*NextX ;	/* Next X Node */
	short		y ;		/* Value of y */
	short		dy ;		/* Delta y */
	int		x ;		/* Start X co-ordinate (in Nibble Fractions) */
	int		m ;		/* dx/dy */
} Edge ;

/* Local Structure used to store Y Buckets */
typedef struct YBucket {
	struct YBucket	*NextY ;	/* Next Y Bucket */
	struct YBucket	*PrevY ;	/* Prev Y Bucket */
	Edge		*NextX ;	/* First X Entry */
	int		y ;		/* Value of y */
} YBucket ;

void	(*PolyBlit)() ;			/* Function to to select appropriate PolyBlit() */

/* -------------------------------------------------------------------------------- */
#if ( GDEBUG & POLYDEBUG )

/* Routine used to print out line blit information */

static void PolyLineP( TGCValues *values, int x1, int y1, int dy, int m )
{
	int x2 ;
	int y2 ; 
	
	values	= values ;			/* Compiler quiet */
	
	if( !dy )	x2 = x1 + m ;
	else		x2 = x1 + m * dy ;

	y2 	= y1 + dy ;
	
	x1 	= FractionstoPixels( x1 ) ;
	x2 	= FractionstoPixels( x2 ) ;

	printf("Line : dy = %3d, m = %6d, (%3d,%3d), (%3d,%3d)\n", dy, m, x1, y1, x2, y2 ) ;
}

/* -------------------------------------------------------------------------------- */

/* Routine used to print out list of edges on completion of input */

static void PolyPrint( char *txt , YBucket *bucket )
{
	Edge 	*edge ;
	char	type ;
	
	while( bucket )
	{
		edge = bucket->NextX ;
		type = 'Y' ;
					
		while( edge )
		{
			printf("%-10s %c : ", txt, type ) ;
			PolyLineP( NULL, edge->x, edge->y, edge->dy, edge->m ) ;
			edge = edge->NextX ;			
			type = 'X' ;
		}
		
		bucket = bucket->NextY ;
	}
}
#endif
/* -------------------------------------------------------------------------------- */

void PolyBlitB( int tlx, int blx, int trx, int brx, int y, int dy, TGCValues *values )
{
	int	li, ri ;	/* Right and left hand side increment (fracs) */
	int	ls, rs ;	/* Left and right hand stride in nibbles */
	int	lx, rx ;	/* Start x for left and right sides */
	int	dx, bc ;	/* Difference in x, value of bc to use for blit */

	/* If the top two are the same decrement the lhs by 1 */
/*	if( trx == tlx ) tlx-- ;	*/
	
	/* Get the difference at the bottom */
	dx = brx - blx ;
	
	/* If there is no margin for rounding add the margin */
	if( dx < dy )
	{
		brx = ( brx + blx + dy ) >> 1 ;
		blx = brx - dy ;
	}

	/* Now calculate out the start positions */
	lx = WholeNibbles( tlx ) ;
	rx = WholeNibbles( trx ) ;

	dx = ScreenBaseinNibbles + YtoNibbAddr( y ) ;
	lx += dx ;
	rx += dx ;

	/* Initialise those parts of the blitter which we need to */
	bc		= InitBlit( blitter, values, TypePolygon ) ;
	blitter->DL 	= dy ;
	blitter->DS 	= lx ;
	blitter->DF 	= GetBlitFractions( tlx ) ;
	blitter->SS 	= rx ;
	blitter->SF 	= GetBlitFractions( trx ) ;
	
	/* Now do the lhs */
	if( blx >= tlx )
	{
		dx = blx - tlx ;
		li = dx / dy ;
		ls = WholeNibbles( li ) ;
		blitter->DT = StrideinNibbles + ls ;
		blitter->DI = GetBlitFractions( li ) ;
	}
	else
	{
		dx = tlx - blx ;
		li = dx / dy ;
		ls = WholeNibbles( li ) ;
		li = GetBlitFractions( li ) ;
		
		if( li )
		{
			blitter->DT = StrideinNibbles - ls - 1 ;
			blitter->DI = BlitFractions - li ;
		}
		else	
		{
			blitter->DT = StrideinNibbles - ls ;
			blitter->DI = li ;
		}
	}
	
	/* And similarly for the rhs ( forcing right ) */
	if( brx >= trx )
	{
		dx = brx - trx ;
		ri = dx / dy ;
		rs = WholeNibbles( ri ) ;
		blitter->ST = StrideinNibbles + rs ;
		blitter->SI = GetBlitFractions( ri ) ;
	}
	else
	{
		dx = trx - brx ;
		ri = dx / dy ;
		rs = WholeNibbles( ri ) ;
		ri = GetBlitFractions( ri ) ;

		if( ri )
		{
			blitter->ST = StrideinNibbles - rs - 1 ;
			blitter->SI = BlitFractions - ri ;
		}
		else	
		{
			blitter->ST = StrideinNibbles - rs ;
			blitter->SI = ri ;
		}
	}

	/* Now tell it to start */
	blitter->BC = bc ;
}

/* -------------------------------------------------------------------------------- */

void PolyBlitT( int tlx, int blx, int trx, int brx, int y, int dy, TGCValues *values )
{
	int	lx, rx ;	/* Right and left start positions */
	int	li, ri ;	/* Right and left hand side increment (fracs) */
	int	ls, rs ;	/* Left and right hand stride in nibbles */
	int	lf, rf ;	/* Left and right Fractions */
	int	dx ;		/* Difference in x */
	char	*addr ;
	
	/* Lift dy up to extra line */
	dy++ ;

	/* Modify co-ordinates to remove any overlaps */
	if( trx == tlx ) trx++ ;/* Increment trx to prevent overlaps when trx == tlx */
	dx = brx - blx ;	/* Get the difference of the bottoms */
	if( dx < dy )		/* Add the two, add the effective error term dy, and average */
	{
		brx = ( brx + blx + dy ) >> 1 ;
		blx = brx - dy ;
	}
	
	/* Calculate lhs */
	if( blx >= tlx )
	{
		dx = blx - tlx ;
		li = dx / dy ;
		ls = WholeBytes( li ) ;
		li = GetBlitFractions( li ) ;
		ls += StrideinBytes ;
	}
	else
	{
		dx = tlx - blx ;
		li = dx / dy ;
		ls = WholeBytes( li ) ;
/*		if( li )	tlx -= ( li - 1 ) ;	*/
		li = GetBlitFractions( li ) ;
		
		if( li )
		{
			ls = StrideinBytes - ls - 1 ;
			li = BlitFractions - li ;
		}
		else	ls = StrideinBytes - ls ;
	}
	
	/* And similarly for the rhs */
	if( brx >= trx )
	{
		dx = brx - trx ;
		ri = dx / dy ;
		rs = WholeBytes( ri ) ;
/*		if( ri )	trx += ( ri - 1 ) ;	*/
		ri = GetBlitFractions( ri ) ;
		rs += StrideinBytes ;
	}
	else
	{
		dx = trx - brx ;
		ri = dx / dy ;
		rs = WholeBytes( ri ) ;
		ri = GetBlitFractions( ri ) ;
		
		if( ri )
		{
			rs = StrideinBytes - rs - 1;
			ri = BlitFractions - ri ;
		}
		else	rs = StrideinBytes - rs ;
	}
		
	/* And finally calculate out the start positions */
	lx = WholeBytes( tlx ) ;
	rx = WholeBytes( trx ) ;
	lf = GetBlitFractions( tlx ) ;
	rf = GetBlitFractions( trx ) ;

	dx = ScreenBaseinBytes + YtoByteAddr( y ) ;
 	lx += dx ;
	rx += dx ;

	while( dy > 0 )
	{
		/* Plot points */
		addr	= (char*)lx ;
		*addr	= values->fg ;
			
		bytblt( addr,addr+1, 0,0, 0,1, rx-lx,1, 0 ) ;

		lx += ls ;
		lf += li ;
		if( lf >= BlitFractions )
		{
			lf -= BlitFractions ;
			lx++ ;
		}

		rx += rs ;
		rf += ri ;
		if( rf >= BlitFractions )
		{
			rf -= BlitFractions ;
			rx++ ;
		}

		dy-- ;
	}
}

/* -------------------------------------------------------------------------------- */

/* RemoveX() removes the given X node from the list (and removes the y bucket if appropriate */

static YBucket *RemoveX( Edge *edge, YBucket *bucket )
{
	Edge	*prev ;
	YBucket *curr = NULL ;

	if( bucket->y == edge->y  &&  bucket->NextX )
	{
		/* Special case if edge is the first one */
		if( bucket->NextX != edge )
		{
			prev = bucket->NextX ;
			while( prev->NextX )
			{
				if( prev->NextX == edge )
				{
					prev->NextX = edge->NextX ;
					break ;
				}
				else	prev = prev->NextX ;
			}
		}
		else	bucket->NextX = edge->NextX ;
	
		edge->NextX = NULL ;
		
		/* If the bucket is empty also remove that */
		if( ! bucket->NextX )
		{
			if( bucket->NextY )
			{
				curr		= bucket->NextY ;
				curr->PrevY	= bucket->PrevY ;
			}
		
			if( bucket->PrevY )
			{
				curr		= bucket->PrevY ;
				curr->NextY	= bucket->NextY ;
			}
		}
		else	curr = bucket ;
	}
	else	curr = bucket ;

	return( curr ) ;
}

/* -------------------------------------------------------------------------------- */

/* InsertX() inserts the given x edge into the given y bucket. The edges are sorted in
 * x order, then slope
 */
 
static void InsertX( Edge *edge, YBucket *bucket )
{
	Edge	*curr = bucket->NextX ;

	/* Reject if the two do not agree */
	if( bucket->y != edge->y )	return ;

	/* See if the edge should be inserted at the head of the list */
	if( edge->x < curr->x )
	{
		edge->NextX 	= bucket->NextX ;
		bucket->NextX 	= edge ;
	}
	else if( edge->x == curr->x  &&  edge->m < curr->m )
	{
		edge->NextX	= bucket->NextX ;
		bucket->NextX	= edge ;
	}
	else
	{
		while( curr->NextX  &&  curr->NextX->x < edge->x )
			curr = curr->NextX ;
			
		while( curr->NextX  &&  curr->NextX->x == edge->x )
		{
			if( curr->NextX->m > edge->m )	break ;
			else				curr = curr->NextX ;
		}
			
		edge->NextX 	= curr->NextX ;
		curr->NextX	= edge ;
	}
}

/* -------------------------------------------------------------------------------- */

/* PolyEvenOdd() is a stub for the more complex polygons */

static void PolyEvenOdd( YBucket *bucket, TGCValues *values )
{
	fputs( "PolyEvenOdd not yet implemented", stderr ) ;
	bucket	= bucket ;
	values	= values ;
}

/* -------------------------------------------------------------------------------- */

/* PolyWinding() is as above */

static void PolyWinding( YBucket *bucket, TGCValues *values )
{
	fputs( "PolyWinding not yet implemented", stderr ) ;
	bucket	= bucket ;
	values	= values ;
}

/* -------------------------------------------------------------------------------- */

/* PolyNonconv() is slightly more complex than a simple convex but still straight forward */

static void PolyNonConv( YBucket *bucket, TGCValues *values )
{
	Edge	*lhs, *rhs, *edge ;
	YBucket	*curr ;
	int	dy, ymin, lx, rx ;
	 
	while( 1 )
	{
		/* Set up left hand side and right hand side */
		lhs = bucket->NextX ;
		
		if( !lhs  ||  !lhs->NextX )	break ;
		else				rhs = lhs->NextX ;

		/* Set up initial dy */
		if( lhs->dy < rhs->dy )	dy = lhs->dy ;
		else			dy = rhs->dy ;

		/* See if there is any interference */
		curr 	= bucket ;
		ymin 	= lhs->y + dy ;
		
		if( curr->NextY ) while( curr->NextY  &&  curr->NextY->y <= ymin )
		{
			curr 	= curr->NextY ;
			dy	= curr->y - lhs->y ;
			lx 	= lhs->x + dy * lhs->m ;
			rx	= rhs->x + dy * rhs->m ;
			edge 	= curr->NextX ;

			/* Note the error term dy (in case rounded) */
			while( edge )
			{
				if( edge->x > lx  &&  edge->x < rx + dy ) break ;
				else	edge = edge->NextX ;
			}
			
			if( edge )	break ;
		}
		else
		{
			lx	= lhs->x + dy * lhs->m ;
			rx	= rhs->x + dy * rhs->m ;
		}

		/* Draw the polygon */
		PolyBlit( lhs->x, lx, rhs->x, rx, lhs->y, dy, values ) ;
		
		/* This time remove the lhs and right hand side from the list, remember next */
		bucket = RemoveX( lhs, bucket ) ;
		bucket = RemoveX( rhs, bucket ) ;

		/* Check if still some left on lhs and/or rhs */
		if( bucket )
		{
			if( dy < lhs->dy )
			{
				lhs->x 	= lx ;
				lhs->y	+= dy ;
				lhs->dy	-= dy ;
				InsertX( lhs, curr ) ;
			}
			
			if( dy < rhs->dy )
			{
				rhs->x 	= rx ;
				rhs->y	+= dy ;
				rhs->dy	-= dy ;
				InsertX( rhs, curr ) ;
			}
		}
		else	break ;
	}
} 	

/* -------------------------------------------------------------------------------- */

/* PolyConvex() is the simplest routine for drawing a completely convex polygon */

static void PolyConvex( YBucket *bucket, TGCValues *values )
{
	Edge	*lhs, *rhs ;
	int	dy, lx, rx ;
	 
	while( 1 )
	{
		/* Set up left hand side and right hand side */
		lhs = bucket->NextX ;
		rhs = lhs->NextX ;

		/* Set up initial dy */
		if( lhs->dy < rhs->dy )	dy = lhs->dy ;
		else			dy = rhs->dy ;

		lx  = lhs->x + lhs->m * dy ;
		rx  = rhs->x + rhs->m * dy ;
		
		/* Draw the polygon */
		PolyBlit( lhs->x, lx, rhs->x, rx, lhs->y, dy, values ) ;
		
		/* Increment bucket */
		if( bucket->NextY )
		{
			bucket = bucket->NextY ;		
			
			/* Check if still some left on lhs and/or rhs (remember slope) */
			if( dy  &&  dy < lhs->dy )
			{
				lhs->x 	= lx ;
				lhs->y	+= dy ;
				lhs->dy	-= dy ;
				InsertX( lhs, bucket ) ;
			}
			else if( dy  &&  dy < rhs->dy )
			{
				rhs->x 	= rx ;
				rhs->y	+= dy ;
				rhs->dy	-= dy ;
				InsertX( rhs, bucket ) ;
			}
		}
		else	break ;
	}
} 	
			
/* -------------------------------------------------------------------------------- */

int TFillPolygon( TDrawable drawable, TGC gc, TPoint *xy, int npoints, int shape, int mode )
{
	YBucket *ybucket,		/* First y bucket allocated */
		*ynext,			/* Next free y bucket */
		*ycurr ;		/* Current y bucket */
	Edge	*xcurr ;		/* Current edge */
	int	i, edges, added,	/* Loop counter, number of edges, flag if added */
		tlx, tly,		/* Top left x and y co-ordinates */
		x1, y1, x2, y2 ;	/* Co-ordinates */
	TPoint	*xy1, *xy2 ;		/* Temporary xy co-ordinates */

	/* Allocate memory for edges */
	ybucket	= malloc( ( sizeof( YBucket ) + sizeof( Edge ) ) * npoints ) ;
	
	/* If OK Initialise data pointers */
	if( ybucket )	
	{
		edges = npoints ;
		ynext = ybucket ;
		xcurr = (Edge*)( ybucket + edges) ;
		added = 0 ;
	}
	else	return( EC_Error | EG_NoMemory ) ;

	/* See if the last point is the same as the first */
	xy1	= xy + edges- 1 ;

	if( xy1->x == xy->x  &&  xy1->y == xy->y )
	{
		edges-- ;
		xy1-- ;
	}

	tlx	= 0 ;	/* Alter these when windows introduced */
	tly	= 0 ;
	
	/* Initialise appropriate x and y */
	x2 	= tlx + PixelstoFractions( xy1->x ) + HalfPixelinFractions ;
	y2 	= tly + xy1->y ;
	xy2	= xy ;
	
	/* For each point calculate out the edge details */
	for( i = edges ; i > 0 ; i-- )
	{
		/* Set up x1, y1, and x2, y2 */
		x1	= x2 ;
		y1	= y2 ;
		x2	= PixelstoFractions( xy2->x ) + HalfPixelinFractions ;
		y2	= xy2->y ;

		if	( mode == Previous )	x2 += x1,	y2 += y1 ;
		else				x2 += tlx, 	y2 += tly ;
		
		xy2++ ;

		/* Find out greatest y, ignoring horizontal edges, and set values */
		if( y1 == y2 )
		{
			edges-- ;
			continue ;
		}

		/* Note that the slope is set to be the nearest (by adding dy/2) */
		if( y1 < y2 )
		{
			xcurr->x	= x1 ;
			xcurr->y	= y1 ;
			xcurr->dy 	= y2 - y1 ;
			xcurr->m	= ( x2 - x1 ) / xcurr->dy ;
		}
		else
		{
			xcurr->x	= x2 ;
			xcurr->y	= y2 ;
			xcurr->dy	= y1 - y2 ;
			xcurr->m	= ( x1 - x2 ) / xcurr->dy ;
		}

		xcurr->NextX	= NULL ;

		/* Add clipping here */
		if( gc->rects == TRUE )	;	/* ClipbyRects (need to handle verticals) */
		
		/* If this is the first one just set the structures */
		if( ! added ) 
		{
			ycurr		= ynext++ ;
			ycurr->NextY	= ycurr->PrevY	= NULL ;
			ycurr->NextX	= xcurr ;
			ycurr->y	= xcurr->y ;
			added		= 1 ;
			xcurr++ ;
		}
		else 
		{
			/* Find the nearest edge with y lower than or equal to xcurr->y */
			if( ycurr->y > xcurr->y )
			{
			 	while( ycurr->y > xcurr->y  &&  ycurr->PrevY )
					ycurr = ycurr->PrevY ;
			}
			else	while( ycurr->y < xcurr->y  &&  ycurr->NextY )
					ycurr = ycurr->NextY ;
			
			/* If one of same y not found insert a new one */
			if( ycurr->y < xcurr->y )
			{
				ynext->NextX	= xcurr ;
				ynext->y	= xcurr->y ;

				ynext->PrevY	= ycurr ;

				if( ycurr->NextY )
				{
					ynext->NextY		= ycurr->NextY ;
					ycurr->NextY->PrevY	= ynext ;
				}
				else	ynext->NextY = NULL ;
				
				ycurr->NextY	= ynext++ ;	/* Remember to increment next */
				xcurr++ ;			/* ditto */
			}
			else if( ycurr->y > xcurr->y )
			{
				ynext->NextX	= xcurr ;
				ynext->y	= xcurr->y ;
				
				ynext->NextY	= ycurr ;
				
				if( ycurr->PrevY )
				{
					ynext->PrevY		= ycurr->PrevY ;
					ycurr->PrevY->NextY	= ynext ;
				}
				else	ynext->PrevY = NULL ;
				
				ycurr->PrevY	= ynext++ ;
				xcurr++ ;
			}
			else
			{
				InsertX( xcurr, ycurr ) ;
				xcurr++ ;
			}
		}
	}

	if( edges > 1 )
	{
		/* Find lowest y */
		while( ycurr->PrevY )	ycurr = ycurr->PrevY ;

#if ( GDEBUG & POLYDEBUG )
		PolyPrint( "Edges", ycurr ) ;
#endif
	
		/* Send it on to the drawing routine */
		if( edges > 3 )
		{
			if( shape == Complex )
			{
				if( gc->values.fill_rule == EvenOddRule )	
					PolyEvenOdd( ycurr, &gc->values ) ;
				else
					PolyWinding( ycurr, &gc->values ) ;
			}
			else if( shape == Nonconvex  )	PolyNonConv( ycurr, &gc->values ) ;
			else				PolyConvex( ycurr, &gc->values ) ;
		}
		else	PolyConvex( ycurr, &gc->values ) ;
	}
	else if( ! edges )	/* A Horizontal line or dot */
	{
		int x ;
		
		x 	= x1 = x2 = xy->x ;
		y1 	= xy->y ;
		xy++ ;
		
		for( i = 1 ; i < npoints ; i++, xy++ )
		{
			if( mode == Previous )	x += xy->x ;
			else			x =  xy->x ;
				
			if	( x < x1 )	x1 = x ;
			else if	( x > x2 )	x2 = x ;
		}

		TDrawLine( drawable, gc, x1, y1, x2, y1 ) ;
	}
	else			/* A Single edge */
	{
		xcurr 	= ycurr->NextX ;
		LineBlit( xcurr->x, xcurr->y, xcurr->dy, xcurr->m, &gc->values ) ;
	}
	
	free( ybucket ) ;
	return( 0 ) ;
}
