/* fastram.c - Various utilities for moving code and data into fast RAM

 */

 

#ifndef T800

#	define T800

#endif



#include <fastram.h>

#include <module.h>

#include <string.h>

#include <stdio.h>

#include <task.h>



/* GetModuleAddr() returns the address of the module structure containing the function

 *	fn		The function we are interested in

 */

 

Module *GetModuleAddr( VoidFnPtr fn )

{

	word *curr_addr = (word*)(*fn) ;

	

	while( curr_addr > (word*)MemStart )

	{

		if ( ((Module*)curr_addr)->Type == T_Module )	break ;

		else						curr_addr-- ;

	}

	

#ifdef MYDEBUG

	if( curr_addr > (word*)MemStart )

	{

		Module *mod = (Module *)curr_addr ;

		printf("GetModuleAddr : %s, Size=%d bytes, Id=%d, MaxData=%dwords\n",

			mod->Name, mod->Size, mod->Id, mod->MaxData ) ;

	}

	else	printf("GetModuleAddr : Error\n" ) ;

#endif



	if( curr_addr <= (word*)MemStart )	

	{

		fprintf( stderr, "GetModuleAddr : Unable to locate module start\n" ) ;

		return( NULL ) ;

	}

	else	return( (Module *)curr_addr ) ;

}

	

/* GetEofFn() returns the address of the end of the function. It is identified by

 * either the end of the module or by the address of the next function in the module

 *	fn		The function we are interested in.

 */



VoidFnPtr GetEofFn( VoidFnPtr fn )

{

	Module 	*curr_module 	= GetModuleAddr( fn ) ;

	word 	*curr_addr 	= (word*)fn ;

	

	if( curr_module )

	{

		word *eof_module = (word*)((byte*)curr_module + curr_module->Size);

		

		while( curr_addr < eof_module )

		{

			if( *curr_addr == (word)T_Proc )	break ;

			else					curr_addr++ ;

		}

	}

	else

	{

		fprintf( stderr, "GetEofFn : Unable to locate end of module\n" ) ;

		curr_addr = NULL ;

	}



#ifdef MYDEBUG

	if ( *curr_addr == (word)T_Proc )

	{

		Proc *next_proc = (Proc*)curr_addr ;

		printf("GetEofFn : next_proc = %s\n", next_proc->Name) ;

	}

	else if ( curr_addr )

	{

		printf("GetEofFn : eof_module = %p\n", curr_addr) ;

	}

	else	printf("GetEofFn : Error\n" ) ;

#endif



	return( curr_addr ) ;

}

	



/* AccelerateFn() copies the applicable function into Fast RAM. The function should not

 * involve any local static variables

 *

 *	mem		Pointer to address of fast memory carrier structure

 *	fastfn		Address of function to store

 *	

 * Returns the the address of the moved function if successful, otherwise NULL

 */



VoidFnPtr AccelerateFn( Carrier **mem, VoidFnPtr fn )

{

	word size ;

		

#ifdef MYDEBUG

	printf("Address of function = %p\n", fn) ;

#endif



	*mem = NULL ;

	

	size = (word)GetEofFn( fn ) - (word)fn ;

	

	if ( size > 0 )

	{

		if ( ( *mem = AllocFast( size, &MyTask->MemPool) ) == NULL )

		{

			fprintf( stderr, 

				"AccelerateFn : %d bytes of fast RAM unavailable\n", size ) ;

			return( fn ) ;

		}

		else

		{

			memcpy( (*mem)->Addr, fn, size ) ;

			return( (*mem)->Addr ) ;

		}

	}

	else 

	{

		fprintf( stderr, "AccelerateFn : Unable to determine function size\n" ) ;

		return( fn ) ;

	}

}



/* AccelerateData() puts the static/global data associated with a module into fast RAM

 * The data which can be put into fast RAM can be found out by using objed -m

 *	mem		A pointer to the carrier structure used for fast memory

 *	fn()		A function within the module we are interested in

 */

 

word AccelerateData( Carrier **mem, VoidFnPtr fn )

{

	Module *mod = GetModuleAddr( fn ) ;

	word return_val = 0 ;

	

	*mem = NULL ;

	

	if ( mod )

	{

		word size = (mod->MaxData +1) * sizeof( word ) ;

		

		if ( ( *mem = AllocFast( size, &MyTask->MemPool) ) == NULL )

		{

			fprintf( stderr,

				"AccelerateData : %d bytes of fast RAM unavailable\n", size ) ;

			return_val = -1 ;

		}

		else

		{

			byte **data_addr = (byte**)(MyTask->ModTab + mod->Id ) ;

			void (*initfn)() ;

			

#ifdef MYDEBUG

			printf("AccelerateData : ModTab = %p, data_addr = %p, fast = %p\n",

					MyTask->ModTab + mod->Id, *data_addr, (*mem)->Addr ) ;

#endif



			memcpy( (*mem)->Addr, *data_addr, size ) ;

			*data_addr = (*mem)->Addr ;



#ifdef MYDEBUG

			printf("About to initialise at %p\n", *(MyTask->ModTab + mod->Id) ) ;

#endif

			initfn = (byte*)(&(mod->Init))+mod->Init ;

			initfn(MyTask->ModTab);

#ifdef DEBUG

			puts("Returned from initialisation routines") ;

#endif

			return_val = 0 ;

		}

	}

	else 

	{

		fprintf( stderr, "AccelerateData : Unable to find module\n" ) ;

		return_val = -1 ;

	}

	

	return( return_val ) ;

}	

